import astropy.units as u
import astropy.wcs as fitswcs
import gwcs
import numpy as np
from numpy.testing import assert_allclose
import pytest

from astropy.nddata import StdDevUncertainty
from astropy.coordinates import SpectralCoord
from astropy.tests.helper import quantity_allclose
from astropy.wcs import WCS

from .conftest import remote_access
from ..spectra import Spectrum


def test_empty_spectrum():
    spec = Spectrum(spectral_axis=[]*u.um,
                      flux=[]*u.Jy)

    assert isinstance(spec.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.size == 0

    assert isinstance(spec.flux, u.Quantity)
    assert spec.flux.size == 0


def test_create_from_arrays():
    spec = Spectrum(spectral_axis=np.arange(50) * u.AA,
                    flux=np.ones(50) * u.Jy)

    assert isinstance(spec.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.size == 50

    assert isinstance(spec.flux, u.Quantity)
    assert spec.flux.size == 50

    # Test creating spectrum with unknown arguments
    with pytest.raises(ValueError):
        spec = Spectrum(wavelength=np.arange(1, 50) * u.nm,
                        flux=np.ones(48) * u.Jy)


def test_create_from_multidimensional_arrays():
    """
    This is a test for a bug that was fixed by #283. It makes sure that
    multidimensional flux arrays are handled properly when creating Spectrum
    objects.
    """

    freqs = np.arange(50) * u.GHz
    flux = np.ones((5, len(freqs))) * u.Jy
    spec = Spectrum(spectral_axis=freqs, flux=flux)

    assert (spec.frequency == freqs).all()
    assert (spec.flux == flux).all()

    # Mis-matched lengths should raise an exception (unless freqs is one longer
    # than flux, in which case it's interpreted as bin edges)
    flux = np.ones((5, len(freqs) - 10)) * u.Jy
    with pytest.raises(ValueError):
        spec = Spectrum(spectral_axis=freqs, flux=flux)


def test_create_from_quantities():
    wav = np.arange(1, 50) * u.nm
    flux = np.ones(49) * u.Jy
    spec = Spectrum(spectral_axis=wav, flux=flux)

    assert isinstance(spec.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.unit == u.nm
    assert spec.spectral_axis.size == 49

    # Mis-matched lengths should raise an exception (unless freqs is one longer
    # than flux, in which case it's interpreted as bin edges)
    with pytest.raises(ValueError):
        spec = Spectrum(spectral_axis=wav, flux=np.ones(47) * u.Jy)


def test_create_implicit_wcs():
    spec = Spectrum(spectral_axis=np.arange(50) * u.AA,
                    flux=np.ones(50) * u.Jy)

    assert isinstance(spec.wcs, gwcs.wcs.WCS)

    pix2world = spec.wcs.pixel_to_world(np.arange(5, 10))

    assert pix2world.size == 5
    assert isinstance(pix2world, np.ndarray)


def test_create_implicit_wcs_with_spectral_unit():
    spec = Spectrum(spectral_axis=np.arange(1, 50) * u.nm,
                    flux=np.ones(49) * u.Jy)

    assert isinstance(spec.wcs, gwcs.wcs.WCS)

    pix2world = spec.wcs.pixel_to_world(np.arange(5, 10))

    assert pix2world.size == 5
    assert isinstance(pix2world, np.ndarray)


def test_create_with_spectral_coord():

    spectral_coord = SpectralCoord(np.arange(5100, 5150) * u.AA, radial_velocity=u.Quantity(1000.0, "km/s"))
    flux = np.ones(50) * u.Jy
    spec = Spectrum(spectral_axis=spectral_coord, flux=flux)

    assert spec.radial_velocity == u.Quantity(1000.0, "km/s")
    assert isinstance(spec.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.size == 50


def test_create_from_cube():

    flux = np.arange(24).reshape([2,3,4])*u.Jy
    wcs_dict = {"CTYPE1": "RA---TAN", "CTYPE2": "DEC--TAN", "CTYPE3": "WAVE-LOG",
                "CRVAL1": 205, "CRVAL2": 27, "CRVAL3": 3.622e-7,
                "CDELT1": -0.0001, "CDELT2": 0.0001, "CDELT3": 8e-11,
                "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0}
    w = WCS(wcs_dict)

    spec = Spectrum(flux=flux, wcs=w)
    spec_axis_from_wcs = (np.exp(np.array([1,2])*w.wcs.cdelt[-1]/w.wcs.crval[-1]) *
                          w.wcs.crval[-1]*spec.spectral_axis.unit)

    assert spec.flux.shape == (2,3,4)
    assert spec.flux[1,2,3] == 23*u.Jy
    assert quantity_allclose(spec.spectral_axis, spec_axis_from_wcs)

    with pytest.raises(ValueError):
        spec2 = Spectrum(flux=flux, wcs=w, move_spectral_axis='Bad string')

    # Test moving spectral axis from first to last
    spec2 = Spectrum(flux=flux, wcs=w, move_spectral_axis='last')
    assert spec2.flux.shape == (4,3,2)
    assert spec2.flux[3,2,1] == 23*u.Jy
    assert quantity_allclose(spec2.spectral_axis, spec_axis_from_wcs)

    # Test moving spectral axis from last to first
    spec3 = Spectrum(flux=spec2.flux, wcs=spec2.wcs, move_spectral_axis='first')
    assert spec3.flux.shape == (2,3,4)
    assert spec3.flux[1,2,3] == 23*u.Jy
    assert quantity_allclose(spec3.spectral_axis, spec_axis_from_wcs)


def test_spectral_axis_conversions():
    # By default the spectral axis units should be set to angstroms
    spec = Spectrum(flux=np.array([26.0, 44.5]) * u.Jy,
                      spectral_axis=np.array([400, 500]) * u.AA)

    assert np.all(spec.spectral_axis == np.array([400, 500]) * u.angstrom)
    assert spec.spectral_axis.unit == u.angstrom

    flux = np.ones(49) * u.Jy
    spec = Spectrum(spectral_axis=np.arange(flux.size) * u.AA, flux=flux)

    assert spec.wavelength.unit == u.AA

    spec = Spectrum(spectral_axis=np.arange(1, 50) * u.nm, flux=flux)

    assert spec.frequency.unit == u.GHz

    with pytest.raises(ValueError):
        spec.velocity

    spec = Spectrum(spectral_axis=np.arange(100, 150) * u.nm, flux=flux)

    new_spec = spec.with_spectral_axis_unit(u.km / u.s, rest_value=125 * u.um,
                                            velocity_convention="relativistic")

    assert new_spec.spectral_axis.unit == u.km / u.s
    assert new_spec.wcs.world_axis_units[0] == "km.s**-1"
    # Make sure meta stored the old WCS correctly
    assert new_spec.meta["original_wcs"].world_axis_units[0] == "nm"
    assert new_spec.meta["original_spectral_axis_unit"] == "nm"

    wcs_dict = {"CTYPE1": "WAVE", "CRVAL1": 3.622e3, "CDELT1": 8e-2,
                "CRPIX1": 0, "CUNIT1": "Angstrom"}
    wcs_spec = Spectrum(flux=flux, wcs=WCS(wcs_dict),
                          meta={'header': wcs_dict.copy()})
    new_spec = wcs_spec.with_spectral_axis_unit(u.km / u.s, rest_value=125 * u.um,
                                                velocity_convention="relativistic")
    new_spec.meta['original_wcs'].wcs.crval = [3.777e-7]
    new_spec.meta['header']['CRVAL1'] = 3777.0

    assert wcs_spec.wcs.wcs.crval[0] == 3.622e-7
    assert wcs_spec.meta['header']['CRVAL1'] == 3622.


def test_spectral_axis_and_flux_conversions():
    """A little bit from both sets of tests."""
    spec = Spectrum(spectral_axis=np.arange(100, 150) * u.nm,
                      flux=np.ones(49) * u.Jy)

    new_spec = spec.with_spectral_axis_and_flux_units(
        u.km / u.s, u.uJy, rest_value=125 * u.um, velocity_convention="relativistic")

    assert new_spec.spectral_axis.unit == u.km/u.s
    assert new_spec.wcs.world_axis_units[0] == "km.s**-1"
    # Make sure meta stored the old WCS correctly
    assert new_spec.meta["original_wcs"].world_axis_units[0] == "nm"
    assert new_spec.meta["original_spectral_axis_unit"] == "nm"
    assert new_spec.flux.unit == u.uJy
    assert_allclose(new_spec.flux.value, 1000000)


def test_spectral_slice():
    spec = Spectrum(spectral_axis=np.linspace(100, 1000, 10) * u.nm,
                    flux=np.ones(10) * u.Jy)
    sliced_spec = spec[300*u.nm:600*u.nm]
    assert np.all(sliced_spec.spectral_axis == [300, 400, 500] * u.nm)

    sliced_spec = spec[300*u.nm:605*u.nm]
    assert np.all(sliced_spec.spectral_axis == [300, 400, 500, 600] * u.nm)

    sliced_spec = spec[:300*u.nm]
    assert np.all(sliced_spec.spectral_axis == [100, 200] * u.nm)

    sliced_spec = spec[800*u.nm:]
    assert np.all(sliced_spec.spectral_axis == [800, 900, 1000] * u.nm)

    # Test higher dimensional slicing
    spec = Spectrum(spectral_axis=np.linspace(100, 1000, 10) * u.nm,
                    flux=np.ones((10, 10)) * u.Jy,
                       spectral_axis_index=1)
    sliced_spec = spec[300*u.nm:600*u.nm]
    assert np.all(sliced_spec.spectral_axis == [300, 400, 500] * u.nm)

    sliced_spec = spec[4:6, 300*u.nm:600*u.nm]
    assert sliced_spec.shape == (2, 3)


@pytest.mark.parametrize('unit', ['micron', 'GHz', 'cm**-1', 'eV'])
def test_spectral_axis_equivalencies(unit):
    """Test that `u.spectral` equivalencies are enabled for `spectral_axis`."""

    spectral_axis=np.array([3400, 5000, 6660]) * u.AA
    spec = Spectrum(flux=np.array([26.0, 30.0, 44.5]) * u.Jy, spectral_axis=spectral_axis)

    new_axis = spectral_axis.to(unit, equivalencies=u.spectral())
    assert u.allclose(spec.spectral_axis.to(unit), new_axis)


def test_redshift():
    spec = Spectrum(flux=np.array([26.0, 30., 44.5]) * u.Jy,
                      spectral_axis=np.array([4000, 6000, 8000]) * u.AA,
                      velocity_convention='optical',
                      rest_value=6000 * u.AA)

    assert u.allclose(spec.velocity, [-99930.8, 0, 99930.8]*u.km/u.s,
                     atol=0.5*u.km/u.s)

    spec = Spectrum(flux=np.array([26.0, 30., 44.5]) * u.Jy,
                      spectral_axis=np.array([4000, 6000, 8000]) * u.AA,
                      velocity_convention='optical',
                      rest_value=6000 * u.AA,
                      redshift= 0.1)

    assert u.allclose(spec.velocity, [-71443.75318854, 28487.0661448, 128417.88547813]*u.km/u.s,
                      atol=0.5*u.km/u.s)

    # -------------------------

    spec = Spectrum(flux=np.array([26.0, 30.0, 44.5]) * u.Jy,
                      spectral_axis=np.array([10.5, 11.0, 11.5]) * u.GHz,
                      velocity_convention='radio',
                      rest_value=11.0 * u.GHz)

    assert u.allclose(spec.velocity, [13626., 0, -13626]*u.km/u.s,
                      atol=1*u.km/u.s)

    spec = Spectrum(flux=np.array([26.0, 30.0, 44.5]) * u.Jy,
                      spectral_axis=np.array([10.5, 11.0, 11.5]) * u.GHz,
                      velocity_convention='radio',
                      rest_value=11.0 * u.GHz,
                      redshift= 0.1)

    assert u.allclose(spec.velocity, [42113.99605389, 28487.0661448 , 14860.13623571]*u.km/u.s,
                      atol=1*u.km/u.s)

    # ------------------------- radial velocity mode

    spec = Spectrum(flux=np.array([26.0, 30., 44.5]) * u.Jy,
                      spectral_axis=np.array([4000, 6000, 8000]) * u.AA,
                      velocity_convention='optical',
                      rest_value=6000 * u.AA)

    assert u.allclose(spec.velocity, [-99930.8, 0.0, 99930.8]*u.km/u.s,
                      atol=0.5*u.km/u.s)

    spec = Spectrum(flux=np.array([26.0, 30., 44.5]) * u.Jy,
                      spectral_axis=np.array([4000, 6000, 8000]) * u.AA,
                      velocity_convention='optical',
                      rest_value=6000 * u.AA,
                      radial_velocity=1000.*u.km/u.s)

    assert u.allclose(spec.velocity, [-98930.8, 1000.0, 100930.8]*u.km/u.s,
                      atol=0.5*u.km/u.s)


def test_flux_unit_conversion():
    # By default the flux units should be set to Jy
    s = Spectrum(flux=np.array([26.0, 44.5]) * u.Jy,
                   spectral_axis=np.array([400, 500]) * u.nm)
    assert np.all(s.flux == np.array([26.0, 44.5]) * u.Jy)
    assert s.flux.unit == u.Jy

    # Simple Unit Conversion
    s = Spectrum(flux=np.array([26.0, 44.5]) * u.Jy, spectral_axis=np.array([400, 500])*u.nm)
    converted_spec = s.with_flux_unit(unit=u.uJy)
    assert ((26.0 * u.Jy).to(u.uJy) == converted_spec.flux[0])

    # Make sure incompatible units raise UnitConversionError
    with pytest.raises(u.UnitConversionError):
        s.with_flux_unit(unit=u.m)

    # Pass custom equivalencies
    s = Spectrum(flux=np.array([26.0, 44.5]) * u.Jy,
                   spectral_axis=np.array([400, 500]) * u.nm)
    eq = [[u.Jy, u.m,
          lambda x: np.full_like(np.array(x), 1000.0, dtype=np.double),
          lambda x: np.full_like(np.array(x), 0.001, dtype=np.double)]]
    converted_spec = s.with_flux_unit(unit=u.m, equivalencies=eq)
    assert 1000.0 * u.m == converted_spec.flux[0]

    # Check if suppressing the unit conversion works
    s = Spectrum(flux=np.array([26.0, 44.5]) * u.Jy, spectral_axis=np.array([400, 500]) * u.nm)
    new_spec = s.with_flux_unit("uJy", suppress_conversion=True)
    assert new_spec.flux[0] == 26.0 * u.uJy


def test_wcs_transformations():
    # Test with a GWCS
    spec = Spectrum(spectral_axis=np.arange(1, 50) * u.nm,
                    flux=np.ones(49) * u.Jy)

    # After spacetelescope/gwcs#457 is merged and released, this can be changed to
    # pix_axis = spec.wcs.world_to_pixel_values(np.arange(20, 30) * u.nm)
    pix_axis = spec.wcs.world_to_pixel(SpectralCoord(np.arange(20, 30) * u.nm))
    disp_axis = spec.wcs.pixel_to_world(np.arange(20, 30))

    assert isinstance(pix_axis, np.ndarray)
    assert isinstance(disp_axis, u.Quantity)

    # Test transform with different unit
    with u.set_enabled_equivalencies(u.spectral()):
        # After spacetelescope/gwcs#457 is merged and released, this can be changed to
        # spec.wcs.world_to_pixel_values(np.arange(20, 30) * u.GHz)
        spec.wcs.world_to_pixel(SpectralCoord(np.arange(6.9e6, 1e7) * u.GHz))

    # Test with a FITS WCS
    my_wcs = fitswcs.WCS(header={'CDELT1': 1, 'CRVAL1': 6562.8, 'CUNIT1': 'Angstrom',
                                 'CTYPE1': 'WAVE', 'RESTFRQ': 1400000000, 'CRPIX1': 25},
                         naxis=1)

    spec = Spectrum(flux=[5,6,7] * u.Jy, wcs=my_wcs)

    pix_axis = spec.wcs.world_to_pixel(20 * u.um)
    disp_axis = spec.wcs.pixel_to_world(np.arange(20, 30))

    assert isinstance(pix_axis, np.ndarray)
    assert isinstance(disp_axis, u.Quantity)

    assert np.allclose(spec.wcs.world_to_pixel(7000*u.AA), [461.2])


def test_create_explicit_fitswcs():
    my_wcs = fitswcs.WCS(header={'CDELT1': 1, 'CRVAL1': 6562.8, 'CUNIT1': 'Angstrom',
                                 'CTYPE1': 'WAVE', 'RESTFRQ': 1400000000, 'CRPIX1': 25})

    spec = Spectrum(flux=[5,6,7] * u.Jy, wcs=my_wcs)
    spec = spec.with_velocity_convention("relativistic")

    assert isinstance(spec.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.unit.is_equivalent(u.AA)

    pix2world = spec.wcs.pixel_to_world(np.arange(3))

    assert pix2world.size == 3

    assert isinstance(spec.wavelength, u.Quantity)
    assert spec.wavelength.size == 3
    assert spec.wavelength.unit == u.AA

    assert isinstance(spec.frequency, u.Quantity)
    assert spec.frequency.size == 3
    assert spec.frequency.unit == u.GHz

    assert isinstance(spec.velocity, u.Quantity)
    assert spec.velocity.size == 3
    assert spec.velocity.unit == u.Unit('km/s')


def test_create_with_uncertainty():
    spec = Spectrum(spectral_axis=np.arange(1, 50) * u.nm,
                    flux=np.ones(49) * u.Jy,
                    uncertainty=StdDevUncertainty(np.ones(49) * 0.1))

    assert isinstance(spec.uncertainty, StdDevUncertainty)
    assert spec.flux.unit == spec.uncertainty.unit

    # If flux and uncertainty are different sizes then raise exception
    wavelengths = np.arange(10) * u.um
    flux= np.ones((3, 4, 10)) * u.Jy
    uncertainty = StdDevUncertainty(np.ones((3, 2, 10)) * u.Jy)

    with pytest.raises(ValueError):
        Spectrum(spectral_axis=wavelengths, flux=flux, uncertainty=uncertainty)


@pytest.mark.parametrize("flux_unit", ["adu", "ct/s", "count"])
def test_flux_unit_io_roundtrip(tmp_path, flux_unit):
    # regression test for https://github.com/astropy/specutils/pull/1018
    fname = str(tmp_path / 'flux_unit_io_roundtrip.fits')
    sp = Spectrum(flux=np.ones(11) * u.Unit(flux_unit),
                    spectral_axis=np.arange(1, 12) * u.Unit('Hz'))
    sp.write(fname, overwrite=True)

    sp_load = Spectrum.read(fname)
    assert sp_load.flux.unit == sp.flux.unit


@pytest.mark.filterwarnings('ignore::astropy.io.fits.verify.VerifyWarning')
@remote_access([{'id': '1481190', 'filename': 'L5g_0355+11_Cruz09.fits'}])
def test_read_linear_solution(remote_data_path):
    spec = Spectrum.read(remote_data_path, format='wcs1d-fits')

    assert isinstance(spec, Spectrum)

    assert isinstance(spec.flux, u.Quantity)
    assert isinstance(spec.spectral_axis, SpectralCoord)

    assert spec.flux.size == spec.data.size
    assert spec.spectral_axis.size == spec.data.size


def test_energy_photon_flux():
    spec = Spectrum(spectral_axis=np.linspace(100, 1000, 10) * u.nm,
                    flux=np.ones(10) * u.Jy)
    assert spec.energy.size == 10
    assert spec.photon_flux.size == 10
    assert spec.photon_flux.unit == u.photon * u.cm**-2 * u.s**-1 * u.nm**-1


def test_flux_nans_propagate_to_mask():
    """Check that indices in input flux with NaNs get propagated to the mask"""
    flux = np.ones(10)
    nan_idx = [0, 3, 5]
    flux[nan_idx] = np.nan
    spec = Spectrum(spectral_axis=np.linspace(100, 1000, 10) * u.nm,
                      flux=flux * u.Jy)
    assert spec.mask[nan_idx].all() == True  # noqa


def test_repr():
    wav = np.linspace(100, 1000, 10) * u.nm
    flux = np.ones(10) * u.Jy
    spec_with_wcs = Spectrum(spectral_axis=wav, flux=flux)
    result = repr(spec_with_wcs)
    assert result.startswith('<Spectrum(flux=<Quantity [')
    assert 'spectral_axis=<SpectralAxis' in result

    spec_with_unc = Spectrum(spectral_axis=wav, flux=flux,
                             uncertainty=StdDevUncertainty(flux))
    result = repr(spec_with_unc)
    assert result.startswith('<Spectrum(flux=<Quantity [')
    assert 'spectral_axis=<SpectralAxis' in result
    assert 'uncertainty=StdDevUncertainty' in result


def test_str():
    wav = np.linspace(100, 1000, 10) * u.nm
    flux = np.ones(10) * u.Jy
    spec = Spectrum(spectral_axis=wav, flux=flux)
    result = str(spec)
    # Sanity check for contents of string representation
    assert result.startswith('Spectrum (length={})'.format(len(spec.flux)))
    lines = result.split('\n')
    flux = spec.flux
    sa = spec.spectral_axis
    assert len(lines) == 3
    assert lines[1].startswith('Flux=')
    assert 'mean={:.5f}'.format(np.nanmean(flux)) in lines[1]
    assert lines[2].startswith('Spectral Axis=')
    assert f'mean={np.nanmean(sa):.5f}' in lines[2]

    # Test string representation with uncertainty
    spec_with_uncertainty = Spectrum(spectral_axis=wav,
                                     flux=flux,
                                     uncertainty=StdDevUncertainty(flux.value))
    result = str(spec_with_uncertainty)
    lines = result.split('\n')
    unc = spec_with_uncertainty.uncertainty
    print(spec_with_uncertainty)
    assert len(lines) == 4
    assert lines[3].startswith('Uncertainty')
    assert f'StdDevUncertainty ([{unc.array[0]:.0f}' in lines[3]

    # Test string representation with multiple flux
    spec_multi_flux = Spectrum(spectral_axis=wav,
                               flux=np.ones((3, 10)) * u.Jy)
    result = str(spec_multi_flux)
    lines = result.split('\n')
    assert len(lines) == 5
    assert lines[1].startswith('Flux=')

    # Test string representation with single-dimensional flux
    spec_single_flux = Spectrum([1] * u.Jy, [0] * u.nm)
    result = str(spec_single_flux)
    assert result == \
"""Spectrum (length=1)
Flux=[1.] Jy,  mean=1.00000 Jy
Spectral Axis=[0.] nm,  mean=0.00000 nm"""


def test_equivalencies():
    """
    Test that after import `u.spectral` equivalencies are not enabled in the global namespace.
    """
    assert u.micron.is_equivalent(u.cm**-1) is False
    assert u.micron.is_equivalent(u.Hz) is False
    assert u.micron.is_equivalent(u.eV) is False
    assert u.Hz.is_equivalent(u.cm**-1) is False
    assert u.Hz.is_equivalent(u.eV) is False


def test_collapse_flux():
    flux = [[2, 4, 6], [0, 8, 12]] * u.Jy
    sa = [100, 200, 300] * u.um
    mask = [[False, True, False], [True, False, False]]
    spec = Spectrum(flux, sa, mask=mask, spectral_axis_index=1)

    assert spec.mean() == 7 * u.Jy
    assert spec.max() == 12 * u.Jy
    assert spec.min() == 2 * u.Jy
    assert spec.sum() == 28 * u.Jy
    assert spec.median() == 7 * u.Jy

    sum_spec = spec.sum(axis = 0)
    assert isinstance(sum_spec, Spectrum)
    assert np.all(sum_spec.flux == [2, 8, 18] * u.Jy)

    mean_spec = spec.mean(axis = 0)
    assert isinstance(mean_spec, Spectrum)
    assert np.all(mean_spec.flux == [2, 8, 9] * u.Jy)

    max_spec = spec.max(axis = 0)
    assert isinstance(max_spec, Spectrum)
    assert np.all(max_spec.flux == [2, 8, 12] * u.Jy)

    min_spec = spec.min(axis = 0)
    assert isinstance(min_spec, Spectrum)
    assert np.all(min_spec.flux == [2, 8, 6] * u.Jy)

    median_spec = spec.mean(axis = 0)
    assert isinstance(median_spec, Spectrum)
    assert np.all(median_spec.flux == [2, 8, 9] * u.Jy)


def test_unsorted_spectral_axis_fails():
    """
    Test that creating a Spectrum fails if spectral axis isn't strictly
    ascending or descending.
    """
    wave = [2, 1, 3] * u.nm
    flux = [5, 5, 5] * u.Jy

    with pytest.raises(ValueError, match='Spectral axis must be strictly increasing or decreasing.'):
        Spectrum(spectral_axis=wave, flux=flux)


def test_spectral_axis_direction():
    """
    Test that the spec1d.spectral_axis_direction attribute correctly reflects
    if the spectral axis in Spectrum is increasing or decreasing.
    """
    flux = [5, 5, 5] * u.Jy

    wave = [1, 2, 3] * u.nm
    spec1d = Spectrum(spectral_axis=wave, flux=flux)
    assert spec1d.spectral_axis_direction == 'increasing'

    wave = [3, 2, 1] * u.nm
    spec1d = Spectrum(spectral_axis=wave, flux=flux)
    assert spec1d.spectral_axis_direction == 'decreasing'



import numpy as np
import astropy.units as u
from astropy import constants as const

from astropy.nddata import StdDevUncertainty
from astropy.modeling import models

from ..spectra.spectrum import Spectrum
from ..analysis import correlation


def test_autocorrelation():
    """
    Test auto correlation
    """
    size = 42

    # Seed np.random so that results are consistent
    np.random.seed(41)

    # Create test spectra
    spec_axis = np.linspace(5000., 5040., num=size) * u.AA
    f1 = np.random.randn(size) * u.Jy
    g1 = models.Gaussian1D(amplitude=30 * u.Jy, mean=5020 * u.AA, stddev=2 * u.AA)
    flux1 = f1 + g1(spec_axis)

    # Observed spectrum must have a rest wavelength value set in.
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=flux1,
                       uncertainty=StdDevUncertainty(np.random.sample(size), unit='Jy'),
                       velocity_convention='optical',
                       rest_value=5020.*u.AA)

    spec2 = Spectrum(spectral_axis=spec_axis,
                       flux=flux1,
                       uncertainty=StdDevUncertainty(np.random.sample(size), unit='Jy'))

    # Get result from correlation
    corr, lag = correlation.template_correlate(spec1, spec2)

    # Check units
    assert corr.unit == u.dimensionless_unscaled
    assert lag.unit == u.km / u.s

    # Check that lags are symmetrical
    midpoint = int(len(lag) / 2)
    np.testing.assert_almost_equal(lag[midpoint+1].value, (-(lag[midpoint-1])).value, 2)

    # Check position of correlation peak.
    maximum = np.argmax(corr)
    assert maximum == midpoint

    # Test that this works without uncertainty
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=flux1,
                       velocity_convention='optical',
                       rest_value=5020.*u.AA)

    corr, lag = correlation.template_correlate(spec1, spec2)

    assert corr.unit == u.dimensionless_unscaled
    assert lag.unit == u.km / u.s


def test_correlation():
    """
    Test correlation when both observed and template spectra have the same wavelength axis
    """
    size = 4000

    # Seed np.random so that results are consistent
    np.random.seed(51)

    # Create test spectra
    spec_axis = np.linspace(4001., 8000., num=size) * u.AA

    # Two narrow Gaussians are offset from each other so
    # as to generate a correlation peak at a expected lag.
    f1 = np.random.randn(size)*0.5 * u.Jy
    f2 = np.random.randn(size)*0.5 * u.Jy

    expected_lag = 10000. * u.km/u.s
    mean2 = 5000. * u.AA
    stdev2 = 30. * u.AA
    mean1 = (1 + expected_lag / const.c.to('km/s')) * mean2
    stdev1 = (1 + expected_lag / const.c.to('km/s')) * stdev2
    rest_value = mean2

    g1 = models.Gaussian1D(amplitude=30 * u.Jy, mean=mean1, stddev=stdev1)
    g2 = models.Gaussian1D(amplitude=30 * u.Jy, mean=mean2, stddev=stdev2)

    flux1 = f1 + g1(spec_axis)
    flux2 = f2 + g2(spec_axis)

    # Observed spectrum must have a rest wavelength value set in.
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=flux1,
                       uncertainty=StdDevUncertainty(np.random.sample(size), unit='Jy'),
                       velocity_convention='optical',
                       rest_value=rest_value)

    spec2 = Spectrum(spectral_axis=spec_axis,
                       flux=flux2,
                       uncertainty=StdDevUncertainty(np.random.sample(size), unit='Jy'))

    # Get result from correlation
    corr, lag = correlation.template_correlate(spec1, spec2)

    # Check units
    assert corr.unit == u.dimensionless_unscaled
    assert lag.unit == u.km / u.s

    # Check position of correlation peak.
    maximum = np.argmax(corr)
    v_fit = _fit_peak(corr, lag, maximum)
    # checks against 1.5 * 10**(-decimal)
    np.testing.assert_almost_equal(v_fit.value, expected_lag.value, -1)


def _create_arrays(size1, size2):
    spec_axis_1 = np.linspace(6050., 6100., num=size1) * u.AA
    spec_axis_2 = np.linspace(6000., 6100., num=size2) * u.AA

    # Create continuum-subtracted flux arrays with noise.
    f1 = np.random.randn(size1) * u.Jy
    f2 = np.random.randn(size2) * u.Jy

    # Two narrow Gaussians are offset from each other so
    # as to generate a correlation peak at a expected lag.
    expected_lag = 1000. * u.km/u.s
    mean2 = 6050. * u.AA
    stdev2 = 5. * u.AA
    mean1 = (1 + expected_lag / const.c.to('km/s')) * mean2
    stdev1 = (1 + expected_lag / const.c.to('km/s')) * stdev2
    rest_value = mean2

    g1 = models.Gaussian1D(amplitude=50 * u.Jy, mean=mean1, stddev=stdev1)
    g2 = models.Gaussian1D(amplitude=50 * u.Jy, mean=mean2, stddev=stdev2)

    flux1 = f1 + g1(spec_axis_1)
    flux2 = f2 + g2(spec_axis_2)

    return spec_axis_1, spec_axis_2, flux1, flux2, expected_lag, rest_value


def _fit_peak(corr, lag, index_peak):

    # Parabolic fit to maximum

    n = 3  # points to the left and right of correlation maximum

    peak_lags = lag[index_peak - n:index_peak + n + 1].value
    peak_vals = corr[index_peak - n:index_peak + n + 1].value
    p = np.polyfit(peak_lags, peak_vals, deg=2)
    roots = np.roots(p)

    return np.mean(roots) * u.km / u.s  # maximum lies at mid point between roots


def test_correlation_zero_padding():
    """
    Test correlation when observed and template spectra have different sizes
    """
    size1 = 51
    size2 = 101

    # Seed np.random so that results are consistent
    np.random.seed(41)

    # Create test spectra
    spec_axis_1, spec_axis_2, flux1, flux2, expected_lag, rest_value = _create_arrays(size1, size2)

    # Observed spectrum must have a rest wavelength value set in.
    # Uncertainty is arbitrary.
    spec1 = Spectrum(spectral_axis=spec_axis_1,
                       flux=flux1,
                       uncertainty=StdDevUncertainty(np.random.sample(size1), unit='Jy'),
                       velocity_convention='optical',
                       rest_value=rest_value)

    spec2 = Spectrum(spectral_axis=spec_axis_2,
                       flux=flux2,
                       uncertainty=StdDevUncertainty(np.random.sample(size2), unit='Jy'))

    # Get result from correlation
    corr, lag = correlation.template_correlate(spec1, spec2)

    # Check units
    assert corr.unit == u.dimensionless_unscaled
    assert lag.unit == u.km / u.s

    # Check that lags are symmetrical
    midpoint = int(len(lag) / 2)
    np.testing.assert_almost_equal(lag[midpoint+1].value, (-(lag[midpoint-1])).value, 2)

    # Check position of correlation peak.
    maximum = np.argmax(corr)
    v_fit = _fit_peak(corr, lag, maximum)
    # checks against 1.5 * 10**(-decimal)
    np.testing.assert_almost_equal(v_fit.value, expected_lag.value, -1)


def test_correlation_random_lines():
    """
    Test correlation when observed and template spectra have different sizes, and
    there are non-correlated features in the observed and template spectra.
    """
    size1 = 51
    size2 = 101

    # Seed np.random so that results are consistent
    np.random.seed(41)

    # Create test spectra
    spec_axis_1, spec_axis_2, flux1, flux2, expected_lag, rest_value = _create_arrays(size1, size2)

    # Add random lines to both spectra to simulate non-correlated spectral features.
    #
    # The more lines we add, and the larger amplitude they have, the larger the error in correlation
    # peak position will be. In this specific case, increasing nlines by 1, or increasing the
    # Gaussian amplitudes by a bit, is enough to make the correlation peak move off position.
    #
    # So we can say generically (as expected anyway) that the presence of non-correlated features
    # in observed and template spectra generates an error in the correlation peak position, and
    # that error will be larger as these non-correlated features become more predominant in the
    # data.
    nlines = 20
    for i in range(nlines):
        mean = (spec_axis_1[-1] - spec_axis_1[0]) * np.random.randn(size1) + spec_axis_1[0]
        g1 = models.Gaussian1D(amplitude=10 * u.Jy, mean=mean, stddev=4 * u.AA)
        flux1 += g1(spec_axis_1)
    for i in range(nlines):
        mean = (spec_axis_2[-1] - spec_axis_2[0]) * np.random.randn(size2) + spec_axis_2[0]
        g2 = models.Gaussian1D(amplitude=5 * u.Jy, mean=mean, stddev=4 * u.AA)
        flux2 += g2(spec_axis_2)

    # Observed spectrum must have a rest wavelength value set in.
    # Uncertainty is arbitrary.
    spec1 = Spectrum(spectral_axis=spec_axis_1,
                       flux=flux1,
                       uncertainty=StdDevUncertainty(np.random.sample(size1), unit='Jy'),
                       velocity_convention='optical',
                       rest_value=rest_value)

    spec2 = Spectrum(spectral_axis=spec_axis_2,
                       flux=flux2,
                       uncertainty=StdDevUncertainty(np.random.sample(size2), unit='Jy'))

    # Get result from correlation
    corr, lag = correlation.template_correlate(spec1, spec2)

    # Check units
    assert corr.unit == u.dimensionless_unscaled
    assert lag.unit == u.km / u.s

    # Check that lags are symmetrical
    midpoint = int(len(lag) / 2)
    np.testing.assert_almost_equal(lag[midpoint+1].value, (-(lag[midpoint-1])).value, 2)

    # Check position of correlation peak.
    maximum = np.argmax(corr)
    v_fit = _fit_peak(corr, lag, maximum)
    # checks against 1.5 * 10**(-decimal)
    np.testing.assert_almost_equal(v_fit.value, expected_lag.value, -1)



import io
from specutils import Spectrum
import specutils.io.default_loaders  # noqa
"""
The purpose of this file is to receive a list of loaders from
specutils.spectrum.read.list_formats(), format that list
into something that can be used by `automodapi`, and then
set it as the __doc__.
"""


def _list_of_loaders():
    # Receive list of loaders
    list_of_loaders = io.StringIO()
    Spectrum.read.list_formats(list_of_loaders)

    # Use the second line (which uses "-" to split the
    # first row from the rest) to create the "=" signs
    # which are used to create the table in .rst files
    split_list = list_of_loaders.getvalue().split("\n")
    line_of_equals = split_list[1].replace("-", "=")
    split_list[1] = line_of_equals

    # Combine elements to create formatted table
    # which can be displayed using `automodapi`
    formatted_table = line_of_equals + "\n"
    for line in split_list:
        formatted_table += line + "\n"
    formatted_table += line_of_equals + "\n"
    return formatted_table


__doc__ = _list_of_loaders()



"""
The core specutils data objects package. This contains the
`~astropy.nddata.NDData`-inherited classes used for storing the spectrum data.
"""
from .spectrum import *  # noqa
from .spectral_region import *  # noqa
from .spectrum_collection import *  # noqa
from .spectrum_list import * # noqa
from .spectral_axis import *  # noqa



import os

from astropy.table import Table

from ...spectra import Spectrum
from ..registers import data_loader
from ..parsing_utils import (generic_spectrum_from_table,
                             spectrum_from_column_mapping)


def identify_ecsv(origin, *args, **kwargs):
    """Check if it's an ECSV file."""
    return (isinstance(args[0], str) and
            os.path.splitext(args[0].lower())[1] == '.ecsv')


@data_loader("ECSV", identifier=identify_ecsv, dtype=Spectrum, priority=-10)
def generic_ecsv(file_name, column_mapping=None, **kwargs):
    """
    Read a spectrum from an ECSV file, using generic_spectrum_from_table_loader()
    to try to figure out which column is which.
    The ECSV columns must have units, as `generic_spectrum_from_table_loader`
    depends on this to determine the meaning of the columns.  For manual
    control over the column to spectrum mapping, use the ASCII loader.

    Parameters
    ----------
    file_name: str
        The path to the ECSV file.
    column_mapping : dict
        A dictionary describing the relation between the ECSV file columns
        and the arguments of the `Spectrum` class, along with unit
        information. The dictionary keys should be the ECSV file column names
        while the values should be a two-tuple where the first element is the
        associated `Spectrum` keyword argument, and the second element is the
        unit for the ECSV file column::

            column_mapping = {'FLUX': ('flux', 'Jy')}

    Returns
    -------
    data: Spectrum
        The spectrum that is represented by the data in this table.
    """
    table = Table.read(file_name, format='ascii.ecsv')

    if column_mapping is None:
        return generic_spectrum_from_table(table)

    return spectrum_from_column_mapping(table, column_mapping)



import warnings

import astropy.units as u
import numpy as np
from astropy.nddata import NDUncertainty, StdDevUncertainty
from astropy.coordinates import SpectralCoord

from .spectrum import Spectrum
from astropy.nddata import NDIOMixin

__all__ = ['SpectrumCollection']


class SpectrumCollection(NDIOMixin):
    """
    A class to represent a heterogeneous set of spectra that are the same length
    but have different spectral axes. Spectra that meet this requirement can be
    stored as multidimensional arrays, and thus can have operations performed
    on them faster than if they are treated as individual
    :class:`~specutils.Spectrum` objects.

    For multidimensional spectra that all have the *same* spectral axis, use a
    :class:`~specutils.Spectrum` with dimension greater than 1.  For a
    collection of spectra that have different shapes, use
    :class:`~specutils.SpectrumList`. For more on this topic, see
    :ref:`specutils-representation-overview`.

    The attributes on this class uses the same names and conventions as
    :class:`~specutils.Spectrum`, allowing some operations to work the same.
    Where this does not work, the user can use standard indexing notation to
    access individual :class:`~specutils.Spectrum` objects.

    Parameters
    ----------
    flux : :class:`astropy.units.Quantity`
        The flux data. The trailing dimension should be the spectral dimension.
    spectral_axis : :class:`astropy.units.Quantity`
        The spectral axes of the spectra (e.g., wavelength).  Must match the
        dimensionality of ``flux``.
    wcs : wcs object or None
        A wcs object (if available) for the collection of spectra.  The object
        must follow standard indexing rules to get a sub-wcs if it is provided.
    uncertainty : :class:`astropy.nddata.NDUncertainty` or ndarray
        The uncertainties associated with each spectrum of the collection. In
        the case that only an n-dimensional quantity or ndaray is provided,
        the uncertainties are assumed to be standard deviations. Must match the
        dimensionality of ``flux``.
    mask : ndarray or None
        The n-dimensional mask information associated with each spectrum. If
        present, must match the dimensionality of ``flux``.
    meta : list
        The list of dictionaries containing meta data to be associated with
        each spectrum in the collection.
    """
    def __init__(self, flux, spectral_axis=None, wcs=None, uncertainty=None,
                 mask=None, meta=None, spectral_axis_index=None):
        # Check for quantity
        if not isinstance(flux, u.Quantity):
            raise u.UnitsError("Flux must be a `Quantity`.")

        if spectral_axis is not None:
            if not isinstance(spectral_axis, u.Quantity):
                raise u.UnitsError("Spectral axis must be a `Quantity`.")
            spectral_axis = SpectralCoord(spectral_axis)

            # Ensure that the input values are the same shape
            if not (flux.shape == spectral_axis.shape):
                raise ValueError("Shape of all data elements must be the same.")

        if uncertainty is not None and uncertainty.array.shape != flux.shape:
            raise ValueError("Uncertainty must be the same shape as flux and "
                            "spectral axis.")

        if mask is not None and mask.shape != flux.shape:
            raise ValueError("Mask must be the same shape as flux and "
                            "spectral axis.")

        # Convert uncertainties to standard deviations if not already defined
        # to be of some type
        if uncertainty is not None and not isinstance(uncertainty, NDUncertainty):
            # If the uncertainties are not provided a unit, raise a warning
            # and use the flux units
            if not isinstance(uncertainty, u.Quantity):
                warnings.warn("No unit associated with uncertainty, assuming"
                              f"flux units of '{flux.unit}'.")
                uncertainty = u.Quantity(uncertainty, unit=flux.unit)

            uncertainty = StdDevUncertainty(uncertainty)

        self._flux = flux
        self._spectral_axis = spectral_axis
        self._spectral_axis_index = spectral_axis_index
        self._wcs = wcs
        self._uncertainty = uncertainty
        self._mask = mask
        self._meta = meta

    def __getitem__(self, key):
        flux = self.flux[key]
        if flux.ndim != 1:
            raise ValueError("Currently only 1D data structures may be "
                             "returned from slice operations.")
        spectral_axis = self.spectral_axis[key]
        uncertainty = None if self.uncertainty is None else self.uncertainty[key]
        wcs = None if self.wcs is None else self.wcs[key]
        mask = None if self.mask is None else self.mask[key]
        if self.meta is None:
            meta = None
        else:
            try:
                meta = self.meta[key]
            except KeyError:
                meta = self.meta

        return Spectrum(flux=flux, spectral_axis=spectral_axis,
                          uncertainty=uncertainty, wcs=wcs, mask=mask,
                          meta=meta)

    @classmethod
    def from_spectra(cls, spectra):
        """
        Create a spectrum collection from a set of individual
        :class:`specutils.Spectrum` objects.

        Parameters
        ----------
        spectra : list, ndarray
            A list of :class:`~specutils.Spectrum` objects to be held in the
            collection. Currently the spectral_axis parameters (e.g. observer,
            radial_velocity) must be the same for each spectrum.
        """
        # Enforce that the shape of each item must be the same
        if not all((x.shape == spectra[0].shape for x in spectra)):
            raise ValueError("Shape of all elements must be the same.")

        # Compose multi-dimensional ndarrays for each property
        flux = u.Quantity([spec.flux for spec in spectra])

        # Check that the spectral parameters are the same for each input
        # spectral_axis and create the multi-dimensional SpectralCoord
        sa = [x.spectral_axis for x in spectra]
        if (not all(x.radial_velocity == sa[0].radial_velocity for x in sa) or
                not all(x.target == sa[0].target for x in sa) or
                not all(x.observer == sa[0].observer for x in sa) or
                not all(x.doppler_convention == sa[0].doppler_convention for
                        x in sa) or
                not all(x.doppler_rest == sa[0].doppler_rest for x in sa)):
            raise ValueError("All input spectral_axis SpectralCoord "
                             "objects must have the same parameters.")
        spectral_axis = SpectralCoord(sa,
                            radial_velocity=sa[0].radial_velocity,
                            doppler_rest=sa[0].doppler_rest,
                            doppler_convention=sa[0].doppler_convention,
                            observer=sa[0].observer,
                            target=sa[0].target)

        spectral_axis_index = spectra[0].spectral_axis_index

        # Check that either all spectra have associated uncertainties, or that
        # none of them do. If only some do, log an error and ignore the
        # uncertainties.
        if (not all((x.uncertainty is None for x in spectra)) and
                any((x.uncertainty is not None for x in spectra)) and
                all((x.uncertainty.uncertainty_type ==
                     spectra[0].uncertainty.uncertainty_type
                     for x in spectra))):
            quncs = u.Quantity([spec.uncertainty.quantity for spec in spectra])
            uncertainty = spectra[0].uncertainty.__class__(quncs)
        else:
            uncertainty = None
            warnings.warn("Not all spectra have associated uncertainties of "
                          "the same type, skipping uncertainties.")

        # Check that either all spectra have associated masks, or that
        # none of them do. If only some do, log an error and ignore the masks.
        if (not all((x.mask is None for x in spectra)) and
                any((x.mask is not None for x in spectra))):
            mask = np.array([spec.mask for spec in spectra])
        else:
            mask = None
            warnings.warn("Not all spectra have associated masks, "
                          "skipping masks.")

        # Store the wcs and meta as lists
        wcs = [spec.wcs for spec in spectra]
        meta = [spec.meta for spec in spectra]

        return cls(flux=flux, spectral_axis=spectral_axis,
                   uncertainty=uncertainty, wcs=wcs, mask=mask, meta=meta,
                   spectral_axis_index=spectral_axis_index)

    @property
    def flux(self):
        """The flux in the spectrum as a `~astropy.units.Quantity`."""
        return self._flux

    @property
    def spectral_axis(self):
        """The spectral axes as a `~astropy.units.Quantity`."""
        return self._spectral_axis

    @property
    def spectral_axis_index(self):
        return self._spectral_axis_index

    @property
    def frequency(self):
        """
        The spectral axis as a frequency `~astropy.units.Quantity` (in GHz).
        """
        return self.spectral_axis.to(u.GHz, u.spectral())

    @property
    def wavelength(self):
        """
        The spectral axis as a wavelength `~astropy.units.Quantity` (in
        Angstroms).
        """
        return self.spectral_axis.to(u.AA, u.spectral())

    @property
    def energy(self):
        """
        The spectral axis as an energy `~astropy.units.Quantity` (in eV).
        """
        return self.spectral_axis.to(u.eV, u.spectral())

    @property
    def wcs(self):
        """The WCS's as an object array"""
        return self._wcs

    @property
    def uncertainty(self):
        """The uncertainty in the spectrum as a `~astropy.units.Quantity`."""
        return self._uncertainty

    @property
    def mask(self):
        """The mask array for the spectrum."""
        return self._mask

    @property
    def meta(self):
        """A dictionary of metadata for theis spectrum collection, or `None`."""
        return self._meta

    @property
    def shape(self):
        """
        The shape of the collection. This is *not* the same as
        the shape of the flux et al., because the trailing (spectral)
        dimension is not included here.
        """
        return self.flux.shape[:-1]

    def __len__(self):
        return self.shape[0]

    @property
    def ndim(self):
        """
        The dimensionality of the collection. This is *not* the same as
        the dimensionality of the flux et al., because the trailing (spectral)
        dimension is not included here.
        """
        return self.flux.ndim - 1

    @property
    def nspectral(self):
        """
        The length of the spectral dimension.
        """
        return self.flux.shape[-1]

    def __repr__(self):
        return """SpectrumCollection(ndim={}, shape={})
    Flux units:          {}
    Spectral axis units: {}
    Uncertainty type:    {}""".format(
            self.ndim, self.shape, self.flux.unit, self.spectral_axis.unit,
            self.uncertainty.uncertainty_type if self.uncertainty is not None else None)



import pytest
import numpy as np

import astropy.units as u
from astropy.tests.helper import quantity_allclose
from astropy.modeling import models
from astropy.nddata.nduncertainty import StdDevUncertainty, VarianceUncertainty, InverseVariance

from ..spectra import Spectrum, SpectralRegion
from ..manipulation import noise_region_uncertainty


def pure_noise_spectrum(amplitude=1*u.mJy):
    np.random.seed(42)

    lamb = np.linspace(4000, 10000, 1000)*u.AA
    flux = np.random.randn(1000) * amplitude
    return Spectrum(spectral_axis=lamb, flux=flux)


def test_noise_region_unc():
    nspec = pure_noise_spectrum()

    region = SpectralRegion(5000*u.AA, 6000*u.AA)
    unc_spec = noise_region_uncertainty(nspec, region)

    assert np.allclose(unc_spec.uncertainty.array, 1, atol=0.1)


@pytest.mark.xfail
def test_noise_region_unc_multi():
    nspec = pure_noise_spectrum()

    multi_region = SpectralRegion([(5000*u.AA, 6000*u.AA), (900*u.pixel, 980*u.pixel)])

    unc_spec = noise_region_uncertainty(nspec, multi_region)

    assert quantity_allclose(unc_spec.uncertainty, 1*u.mJy, atol=.01)  # this is a guess at the atol... need to finalize with real data


def test_noise_estimate_uncertainty():
    # an alternate test with a more complicated input spectrum

    np.random.seed(42)

    # Create values for spectrum.
    frequencies = np.linspace(1, 100, 10000) * u.um
    g = models.Gaussian1D(amplitude=1*u.Jy, mean=10*u.um, stddev=1*u.um)
    noise = np.random.normal(0., 0.01, frequencies.shape) * u.Jy
    flux = g(frequencies) + noise
    spectrum = Spectrum(spectral_axis=frequencies, flux=flux)

    # Assign uncertainties using the default standard deviation
    spectral_region = SpectralRegion(50*u.um, 80*u.um)
    spectrum_with_uncertainty = noise_region_uncertainty(spectrum, spectral_region)

    indices = np.nonzero((frequencies >= 50*u.um) & (frequencies <= 80*u.um))
    expected_uncertainty = np.std(flux[indices])*np.ones(len(frequencies))

    assert quantity_allclose(spectrum_with_uncertainty.uncertainty.array,
                             expected_uncertainty.value)
    assert isinstance(spectrum_with_uncertainty.uncertainty, StdDevUncertainty)

    # Same idea, but now with variance.
    spectrum_with_uncertainty = noise_region_uncertainty(spectrum, spectral_region, np.var)

    indices = np.nonzero((frequencies >= 50*u.um) & (frequencies <= 80*u.um))
    expected_uncertainty = np.var(flux[indices])*np.ones(len(frequencies))

    assert quantity_allclose(spectrum_with_uncertainty.uncertainty.array,
                             expected_uncertainty.value)
    assert isinstance(spectrum_with_uncertainty.uncertainty, VarianceUncertainty)

    # Same idea, but now with inverse variance.
    spectrum_with_uncertainty = noise_region_uncertainty(spectrum, spectral_region,
                                                         lambda x: 1/np.var(x))

    indices = np.nonzero((frequencies >= 50*u.um) & (frequencies <= 80*u.um))
    expected_uncertainty = 1/np.var(flux[indices])*np.ones(len(frequencies))

    assert quantity_allclose(spectrum_with_uncertainty.uncertainty.array,
                             expected_uncertainty.value)
    assert isinstance(spectrum_with_uncertainty.uncertainty, InverseVariance)

    # Now try with something that does not return Std, Var or IVar type of noise estimation
    with pytest.raises(ValueError):
        noise_region_uncertainty(spectrum, spectral_region, lambda x: np.std(x)**3)



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

from copy import copy

import numpy as np
import pytest
from astropy import units as u
from astropy.modeling import models

from specutils.spectra import Spectrum

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


def pytest_configure(config):

    if ASTROPY_HEADER:
        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['gwcs'] = 'gwcs'
        PYTEST_HEADER_MODULES['asdf'] = 'asdf'
        PYTEST_HEADER_MODULES['asdf-astropy'] = 'asdf_astropy'
        PYTEST_HEADER_MODULES['stdatamodels'] = 'stdatamodels'
        PYTEST_HEADER_MODULES['ndcube'] = 'ndcube'
        PYTEST_HEADER_MODULES['spectral-cube'] = 'spectral_cube'

        from specutils import __version__
        TESTED_VERSIONS['specutils'] = __version__


class SpectraExamples:
    """
    The ``SpectralExamples`` class is a *container class* that has
    several examples of simple spectra that are to be used in the tests
    (e.g., arithmetic tests, smoothing tests etc).

    The purpose of this being a test class instead of using a `Spectrum`
    directly is that it contains both the `Spectrum` object and the flux
    that was used to *create* the Spectrum.  That's for tests that ensure
    the simpler operations just on the flux arrays are carried through to
    the `Spectrum` operations.

    Each of the spectra are created from a base noise-less spectrum
    constructed from 4 Gaussians and a ramp. Then three example spectra
    are created, and then gaussian random noise is added.

        1. s1_um_mJy_e1 - 4 Gaussians + ramp with one instantion of noise
                          dispersion: um, flux: mJy

        2. s1_um_mJy_e2 - same as 1, but with a different instance of noise
                          dispersion: um, flux: mJy

        3. s1_AA_mJy_e3 - same as 1, but with a third instance of noise
                          dispersion: Angstroms, flux: mJy

        4. s1_AA_nJy_e3 - same as 1, but with a fourth instance of noise
                          dispersion: Angstroms, flux: nJy

        5. s1_um_mJy_e1_masked - same as 1, but with a random set of pixels
                                 masked.

        6. s1_um_mJy_e1_desc - same as 1, but with the spectral axis in
                               descending rather than ascending order.
    """

    def __init__(self):

        #
        # Create the base wavelengths and flux
        #

        self.wavelengths_um = np.linspace(0.4, 1.05, 100)

        g1 = models.Gaussian1D(amplitude=2000, mean=0.56, stddev=0.01)
        g2 = models.Gaussian1D(amplitude=500, mean=0.62, stddev=0.02)
        g3 = models.Gaussian1D(amplitude=-400, mean=0.80, stddev=0.02)
        g4 = models.Gaussian1D(amplitude=-350, mean=0.52, stddev=0.01)
        ramp = models.Linear1D(slope=300, intercept=0.0)

        self.base_flux = (g1(self.wavelengths_um) + g2(self.wavelengths_um) +
                          g3(self.wavelengths_um) + g4(self.wavelengths_um) +
                          ramp(self.wavelengths_um) + 1000)

        #
        # Initialize the seed so the random numbers are not quite as random
        #

        np.random.seed(42)

        #
        # Create two spectra with the only difference in the instance of noise
        #

        self._flux_e1 = self.base_flux + 400 * np.random.random(self.base_flux.shape)
        self._s1_um_mJy_e1 = Spectrum(spectral_axis=self.wavelengths_um * u.um,
                                        flux=self._flux_e1 * u.mJy)

        self._flux_e2 = self.base_flux + 400 * np.random.random(self.base_flux.shape)
        self._s1_um_mJy_e2 = Spectrum(spectral_axis=self.wavelengths_um * u.um,
                                        flux=self._flux_e2 * u.mJy)

        #
        # Create one spectrum with the same flux but in angstrom units
        #

        self.wavelengths_AA = self.wavelengths_um * 10000
        self._s1_AA_mJy_e3 = Spectrum(spectral_axis=self.wavelengths_AA * u.AA,
                                        flux=self._flux_e1 * u.mJy)

        #
        # Create one spectrum with the same flux but in angstrom units and nJy
        #

        self._flux_e4 = (self.base_flux + 400 * np.random.random(self.base_flux.shape)) * 1000000
        self._s1_AA_nJy_e4 = Spectrum(spectral_axis=self.wavelengths_AA * u.AA,
                                        flux=self._flux_e4 * u.nJy)

        #
        # Create one spectrum like 1 but with a mask
        #
        self._s1_um_mJy_e1_masked = copy(self._s1_um_mJy_e1)  # SHALLOW copy - the data are shared with the above non-masked case  # noqa
        self._s1_um_mJy_e1_masked.mask = (np.random.randn(*self.base_flux.shape) + 1) > 0

        # Create a spectrum like 1, but with descending spectral axis
        self._s1_um_mJy_e1_desc = Spectrum(spectral_axis=self.wavelengths_um[::-1] * u.um,
                                             flux=self._flux_e1[::-1] * u.mJy)

    @property
    def s1_um_mJy_e1(self):
        return self._s1_um_mJy_e1

    @property
    def s1_um_mJy_e1_flux(self):
        return self._flux_e1

    @property
    def s1_um_mJy_e2(self):
        return self._s1_um_mJy_e2

    @property
    def s1_um_mJy_e2_flux(self):
        return self._flux_e2

    @property
    def s1_AA_mJy_e3(self):
        return self._s1_AA_mJy_e3

    @property
    def s1_AA_mJy_e3_flux(self):
        return self._flux_e1

    @property
    def s1_AA_nJy_e4(self):
        return self._s1_AA_nJy_e4

    @property
    def s1_AA_nJy_e4_flux(self):
        return self._flux_e4

    @property
    def s1_um_mJy_e1_masked(self):
        return self._s1_um_mJy_e1_masked

    @property
    def s1_um_mJy_e1_desc(self):
        return self._s1_um_mJy_e1_desc


@pytest.fixture
def simulated_spectra():
    """
    The method will be called as a fixture to tests.

    Parameters
    ----------
    N/A

    Return
    ------
    ``SpectralExamples``
        An instance of the SpectraExamples class.

    Examples
    --------
    This fixture can be used in a test as:

    ```
    from .spectral_examples import spectral_examples

    def test_add_spectra(spectral_examples):

        # Get the numpy array of data
        flux1 = define_spectra.s1_um_mJy_e1_flux
        flux2 = define_spectra.s1_um_mJy_e2_flux
        flux3 = flux1 + flux2

        # Calculate using the spectrum/nddata code
        spec3 = define_spectra.s1_um_mJy_e1 + define_spectra.s1_um_mJy_e2

        assert np.allclose(spec3.flux.value, flux3)
    ```

    """
    return SpectraExamples()



import astropy.units as u
import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, InverseVariance

from ..spectra.spectrum import Spectrum
from ..spectra.spectrum_collection import SpectrumCollection
from ..analysis import template_comparison
from astropy.tests.helper import quantity_allclose


def test_template_match_no_overlap():
    """
    Test template_match when both observed and template spectra have no overlap on the wavelength axis.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    spec_axis_no_overlap = np.linspace(52, 103, 50) * u.AA

    spec = Spectrum(spectral_axis=spec_axis,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec1 = Spectrum(spectral_axis=spec_axis_no_overlap,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Get result from template_match
    with pytest.warns(UserWarning, match="Template spectrum has no overlap with observed spectrum"):
        tm_result = template_comparison.template_match(spec, spec1)

    assert np.isnan(tm_result[3])

    assert tm_result[0] is None


def test_template_match_minimal_overlap():
    """
    Test template_match when both observed and template spectra have minimal overlap on the wavelength axis.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    spec_axis_min_overlap = np.linspace(50, 100, 50) * u.AA

    spec_axis[49] = 51.0 * u.AA
    spec_axis_min_overlap[0] = 51.0 * u.AA

    spec = Spectrum(spectral_axis=spec_axis,
                      flux=abs(np.random.randn(50)) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec1 = Spectrum(spectral_axis=spec_axis_min_overlap,
                       flux=abs(np.random.randn(50)) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec1)
    np.testing.assert_almost_equal(tm_result[3], 0)

    assert len(tm_result[0].flux) == 50
    assert quantity_allclose(tm_result[0].flux, spec1.flux*4.17, atol=0.01*u.Jy)


def test_template_match_spectrum():
    """
    Test template_match when both observed and template spectra have the same wavelength axis.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec1)

    # Create new spectrum for comparison
    spec_result = Spectrum(spectral_axis=spec_axis,
                             flux=spec1.flux * template_comparison._normalize_for_template_matching(spec, spec1))

    assert quantity_allclose(tm_result[0].flux, spec_result.flux, atol=0.01*u.Jy)
    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)


def test_template_match_with_resample():
    """
    Test template_match when both observed and template spectra have different wavelength axis using resampling.
    """
    np.random.seed(42)

    # Create test spectra
    spec_axis1 = np.linspace(0, 50, 50) * u.AA
    spec_axis2 = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis1,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec1 = Spectrum(spectral_axis=spec_axis2,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec1)

    # Create new spectrum for comparison
    spec_result = Spectrum(spectral_axis=spec_axis1,
                             flux=spec1.flux * template_comparison._normalize_for_template_matching(spec, spec1))

    assert quantity_allclose(tm_result[0].flux, spec_result.flux, atol=0.01*u.Jy)
    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)


def test_template_match_list():
    """
    Test template_match when template spectra are in a list.
    """
    np.random.seed(42)

    # Create test spectra
    spec_axis1 = np.linspace(0, 50, 50) * u.AA
    spec_axis2 = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis1,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec1 = Spectrum(spectral_axis=spec_axis2,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))
    spec2 = Spectrum(spectral_axis=spec_axis2,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Combine spectra into list
    template_list = [spec1, spec2]

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, template_list)

    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)

    # make sure that multiple template spectra will create a list of
    # chi2 values, one per template.
    assert len(tm_result) == 5


@pytest.mark.filterwarnings('ignore:Not all spectra have associated masks')
def test_template_match_spectrum_collection():
    """
    Test template_match when template spectra are in a SpectrumCollection object.
    """
    np.random.seed(42)

    # Create test spectra
    spec_axis1 = np.linspace(0, 50, 50) * u.AA
    spec_axis2 = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis1,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50)))
    spec1 = Spectrum(spectral_axis=spec_axis2,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50)))
    spec2 = Spectrum(spectral_axis=spec_axis2,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50)))

    # Combine spectra into SpectrumCollection object
    spec_coll = SpectrumCollection.from_spectra([spec1, spec2])

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec_coll)

    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)


def test_template_match_multidim_spectrum():
    """
    Test template matching with a multi-dimensional Spectrum object.
    """
    np.random.seed(42)

    # Create test spectra
    spec_axis1 = np.linspace(0, 50, 50) * u.AA
    spec_axis2 = np.linspace(0, 50, 50) * u.AA

    spec = Spectrum(spectral_axis=spec_axis1,
                      flux=np.random.sample(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50)))
    multidim_spec = Spectrum(spectral_axis=spec_axis2,
                               flux=np.random.sample((2, 50)) * u.Jy,
                               uncertainty=StdDevUncertainty(np.random.sample((2, 50))))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, multidim_spec)

    np.testing.assert_almost_equal(tm_result[3], 250.26870401777543)


def test_template_unknown_redshift():
    """
    Test template redshift when redshift is unknown.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    perm_flux = np.random.randn(50) * u.Jy

    redshift = 2.5

    # Observed spectrum
    spec = Spectrum(spectral_axis=spec_axis * (1+redshift),
                      flux=perm_flux,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Template spectrum
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=perm_flux,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Test redshift parameters
    min_redshift = .5
    max_redshift = 5.5
    delta_redshift = .25
    redshift_trial_values = np.arange(min_redshift, max_redshift, delta_redshift)

    tr_result = template_comparison.template_redshift(observed_spectrum=spec, template_spectrum=spec1,
                                                      redshift=redshift_trial_values)

    assert len(tr_result) == 5
    assert tr_result[1] == 2.5


def test_template_redshift_with_one_template_spectrum_in_match():
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    perm_flux = np.random.randn(50) * u.Jy

    # Test redshift
    redshift = 3

    # Observed spectrum
    spec = Spectrum(spectral_axis=spec_axis * (1+redshift),
                      flux=perm_flux,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Template spectrum
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=perm_flux,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Test redshift parameters
    min_redshift = .5
    max_redshift = 5.5
    delta_redshift = .25
    redshift_trial_values = np.arange(min_redshift, max_redshift+delta_redshift, delta_redshift)

    tm_result = template_comparison.template_match(observed_spectrum=spec,
                                                   spectral_templates=spec1,
                                                   resample_method="flux_conserving",
                                                   redshift=redshift_trial_values)

    assert len(tm_result) == 5
    np.testing.assert_almost_equal(tm_result[1], redshift)
    np.testing.assert_almost_equal(tm_result[3], 1.9062409482056814e-31)


@pytest.mark.filterwarnings('ignore:Not all spectra have associated masks')
def test_template_redshift_with_multiple_template_spectra_in_match():
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    perm_flux = np.random.randn(50) * u.Jy

    # Test redshift
    redshift = 3

    # Observed spectrum
    spec = Spectrum(spectral_axis=spec_axis * (1+redshift),
                      flux=perm_flux,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Template spectrum
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50)))
    spec2 = Spectrum(spectral_axis=spec_axis,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50)))

    # Combine spectra into SpectrumCollection object
    spec_coll = SpectrumCollection.from_spectra([spec1, spec2])

    # Test redshift parameters
    min_redshift = .5
    max_redshift = 5.5
    delta_redshift = .25
    redshift_trial_values = np.arange(min_redshift, max_redshift+delta_redshift, delta_redshift)

    tm_result = template_comparison.template_match(observed_spectrum=spec,
                                                   spectral_templates=spec_coll,
                                                   resample_method="flux_conserving",
                                                   redshift=redshift_trial_values)
    assert len(tm_result) == 5
    # TODO: Determine cause of and fix failing assert
    # np.testing.assert_almost_equal(tm_result[1], redshift)

    np.testing.assert_almost_equal(tm_result[3], 1172.1446112796834)

    # When a spectrum collection is matched with a redshift
    # grid, a list-of-lists is returned with the trial chi2
    # values computed for every combination redshift-template.
    # The external list spans the templates in the collection,
    # while each internal list contains all chi2 values
    # for a given template.
    assert len(tm_result[4]) == 2


def test_template_known_redshift():
    """
    Test template match when the redshift is known.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    perm_flux = np.random.randn(50) * u.Jy

    redshift = 3

    # Observed spectrum
    spec = Spectrum(spectral_axis=spec_axis * (1+redshift),
                      flux=perm_flux,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    # Template spectrum
    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=perm_flux,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    tm_result = template_comparison.template_match(observed_spectrum=spec,
                                                   spectral_templates=spec1,
                                                   resample_method="flux_conserving",
                                                   redshift=redshift)
    assert len(tm_result) == 5
    np.testing.assert_almost_equal(tm_result[1], redshift)
    np.testing.assert_almost_equal(tm_result[3], 1.9062409482056814e-31)


def test_template_match_variance():
    """
    Test template_match when both observed and template spectra have the same wavelength axis.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=VarianceUncertainty(np.random.sample(50)**2, unit='Jy2'))

    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=VarianceUncertainty(np.random.sample(50)**2, unit='Jy2'))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec1)

    # Create new spectrum for comparison
    spec_result = Spectrum(spectral_axis=spec_axis,
                             flux=spec1.flux * template_comparison._normalize_for_template_matching(spec, spec1))

    assert quantity_allclose(tm_result[0].flux, spec_result.flux, atol=0.01*u.Jy)
    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)


def test_template_match_inverse_variance():
    """
    Test template_match when both observed and template spectra have the same wavelength axis.
    """
    # Seed np.random so that results are consistent
    np.random.seed(42)

    # Create test spectra
    spec_axis = np.linspace(0, 50, 50) * u.AA
    spec = Spectrum(spectral_axis=spec_axis,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=InverseVariance(1/np.random.sample(50)**2, unit='1 / Jy2'))

    spec1 = Spectrum(spectral_axis=spec_axis,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=InverseVariance(1/np.random.sample(50)**2, unit='1 / Jy2'))

    # Get result from template_match
    tm_result = template_comparison.template_match(spec, spec1)

    # Create new spectrum for comparison
    spec_result = Spectrum(spectral_axis=spec_axis,
                             flux=spec1.flux * template_comparison._normalize_for_template_matching(spec, spec1))

    assert quantity_allclose(tm_result[0].flux, spec_result.flux, atol=0.01*u.Jy)
    np.testing.assert_almost_equal(tm_result[3], 40093.28353756253)



"""
Loader for SDSS spectrum files: spec_, spSpec_, and spPlate_.

.. _spec: https://data.sdss.org/datamodel/files/BOSS_SPECTRO_REDUX/RUN2D/spectra/PLATE4/spec.html
.. _spSpec: http://classic.sdss.org/dr7/dm/flatFiles/spSpec.php
.. _spPlate: https://data.sdss.org/datamodel/files/BOSS_SPECTRO_REDUX/RUN2D/PLATE4/spPlate.html
"""
import re

import numpy as np
from astropy.io import fits
from astropy.nddata import InverseVariance, StdDevUncertainty
from astropy.table import Table
from astropy.units import Unit
from astropy.wcs import WCS

from ...spectra import Spectrum
from ..parsing_utils import read_fileobj_or_hdulist
from ..registers import data_loader

__all__ = ['spec_identify', 'spSpec_identify', 'spPlate_identify',
           'spec_loader', 'spSpec_loader', 'spPlate_loader']

#
# These are reserved for future use.
#
_spSpec_pattern = re.compile(r'spSpec-\d{5}-\d{4}-\d{3}\.fit')
_spec_pattern = re.compile(r'spec-\d{4,5}-\d{5}-\d{4}\.fits')
_spPlate_pattern = re.compile(r'spPlate-\d{4,5}-\d{5}\.fits')


def _sdss_wcs_to_log_wcs(old_wcs):
    """
    The WCS in the SDSS files does not appear to follow the WCS standard - it
    claims to be linear, but is logarithmic in base-10.
    The wavelength is given by:
    λ = 10^(w0 + w1 * i)
    with i being the pixel index starting from 0. This formula is documented at
    https://classic.sdss.org/dr7/products/spectra/read_spSpec.php and appears to
    be the same across all SDSS-I and SDSS-II data releases (replace dr7 with
    dr<x> in the above URL).

    The FITS standard uses a natural log with a sightly different formulation,
    see WCS Paper 3 (which discusses spectral WCS).

    This function does the conversion from the SDSS WCS to FITS WCS.
    """
    w0 = old_wcs.wcs.crval[0]
    w1 = old_wcs.wcs.cd[0, 0]
    crval = 10**w0
    cdelt = crval * w1 * np.log(10)
    cunit = Unit("Angstrom")
    ctype = "WAVE-LOG"

    w = WCS(naxis=1)
    w.wcs.crval[0] = crval
    w.wcs.cdelt[0] = cdelt
    w.wcs.ctype[0] = ctype
    w.wcs.cunit[0] = cunit
    w.wcs.set()

    return w


def spec_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS and has SDSS-III/IV spec type
    BINTABLE in first extension. This is used for Astropy I/O Registry.
    """
    # Test if fits has extension of type BinTable and check for spec-specific keys
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (
            hdulist[0].header.get("TELESCOP") == "SDSS 2.5-M"
            and hdulist[0].header["VERS2D"].startswith("v5")
            and hdulist[0].header.get("FIBERID", 0) > 0
            and len(hdulist) > 1
            and (isinstance(hdulist[1], fits.BinTableHDU) and hdulist[1].header.get("TTYPE3").lower() == "ivar")
        )


def spSpec_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS with SDSS-I/II spSpec tyamepe data.
    This is used for Astropy I/O Registry.
    """
    # Test telescope keyword and check if primary HDU contains data
    # consistent with spSpec format
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (hdulist[0].header.get("TELESCOP") == "SDSS 2.5-M"
                and hdulist[0].header.get("FIBERID", 0) > 0
                and isinstance(hdulist[0].data, np.ndarray)
                and hdulist[0].data.shape[0] == 5)


def spPlate_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS with SDSS spPlate fibre spectral data.
    This is used for Astropy I/O Registry.
    """
    # Test telescope keyword and check if primary HDU contains data
    # consistent with spSpec format
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (hdulist[0].header.get("TELESCOP") == "SDSS 2.5-M"
                and hdulist[0].header.get("FIBERID", 0) <= 0
                and isinstance(hdulist[0].data, np.ndarray)
                and hdulist[0].data.shape[0] > 5)


@data_loader(
    label="SDSS-III/IV spec",
    identifier=spec_identify,
    extensions=["fits"],
    priority=10,
)
def spec_loader(file_obj, **kwargs):
    """
    Loader for SDSS-III/IV optical spectrum "spec" files.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    data: Spectrum
        The spectrum that is represented by the 'loglam' (wavelength) and 'flux'
        data columns in the BINTABLE extension of the FITS `file_obj`.
    """
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        meta = {"header": header}

        bunit = header.get("BUNIT", "1e-17 erg / (Angstrom cm2 s)")
        if "Ang" in bunit and "strom" not in bunit:
            bunit = bunit.replace("Ang", "Angstrom")
        flux_unit = Unit(bunit)

        # spectrum is in HDU 1
        flux = hdulist[1].data["flux"] * flux_unit

        uncertainty = InverseVariance(hdulist[1].data["ivar"] / flux_unit**2)

        dispersion = 10**hdulist[1].data["loglam"]
        dispersion_unit = Unit("Angstrom")

        mask = hdulist[1].data["and_mask"] != 0

    return Spectrum(
        flux=flux,
        spectral_axis=dispersion * dispersion_unit,
        uncertainty=uncertainty,
        meta=meta,
        mask=mask,
    )


@data_loader(
    label="SDSS-I/II spSpec",
    identifier=spSpec_identify,
    extensions=["fit", "fits"],
    priority=10,
)
def spSpec_loader(file_obj, **kwargs):
    """
    Loader for SDSS-I/II spSpec files.

    The content of these files is documented at
    https://classic.sdss.org/dr7/dm/flatFiles/spSpec.php, with instructions for
    reading them at
    https://classic.sdss.org/dr7/products/spectra/read_spSpec.php.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
           FITS file name, object (provided from name by Astropy I/O Registry),
           or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    data: Spectrum
        The spectrum that is represented by the wavelength solution from the
        header WCS and data array of the primary HDU.
    """
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        # name = header.get('NAME')
        meta = {"header": header}
        wcs = WCS(header).dropaxis(1)

        bunit = header.get("BUNIT", "1e-17 erg / (Angstrom cm2 s)")
        # fix mutilated flux unit
        bunit = bunit.replace("/cm/s/Ang", "/ (Angstrom cm2 s)")
        if "Ang" in bunit and "strom" not in bunit:
            bunit = bunit.replace("Ang", "Angstrom")
        flux_unit = Unit(bunit)
        flux = hdulist[0].data[0, :] * flux_unit

        uncertainty = StdDevUncertainty(hdulist[0].data[2, :] * flux_unit)

        # Fix the WCS if it is claimed to be linear
        if header.get("DC-Flag", 1) == 1:
            fixed_wcs = _sdss_wcs_to_log_wcs(wcs)
        else:
            fixed_wcs = wcs

        mask = hdulist[0].data[3, :] != 0

    return Spectrum(flux=flux,
                      wcs=fixed_wcs,
                      uncertainty=uncertainty,
                      meta=meta,
                      mask=mask)


@data_loader(label="SDSS spPlate",
             identifier=spPlate_identify,
             extensions=["fits"])
def spPlate_loader(file_obj, limit=None, **kwargs):
    """
    Loader for SDSS spPlate files, reading flux spectra from all fibres into single array.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
           FITS file name, object (provided from name by Astropy I/O Registry),
           or HDUList (as resulting from astropy.io.fits.open()).

    limit : :class:`int`, optional
        If set, only return the first `limit` spectra in `flux` array.

    Returns
    -------
    Spectrum
        The spectra represented by the wavelength solution from the header WCS
        and the data array of the primary HDU (typically 640 along dimension 1).
    """
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        meta = {"header": header}
        wcs = WCS(header).dropaxis(1)
        if limit is None:
            limit = header["NAXIS2"]

        bunit = header.get("BUNIT", "1e-17 erg / (Angstrom cm2 s)")
        if "Ang" in bunit and "strom" not in bunit:
            bunit = bunit.replace("Ang", "Angstrom")
        flux_unit = Unit(bunit)
        flux = hdulist[0].data[0:limit, :] * flux_unit
        uncertainty = InverseVariance(hdulist[1].data[0:limit, :] /
                                      flux_unit**2)

        # Fix the WCS if it is claimed to be linear
        if header.get("DC-Flag", 1) == 1:
            fixed_wcs = _sdss_wcs_to_log_wcs(wcs)
        else:
            fixed_wcs = wcs

        mask = hdulist[2].data[0:limit, :] != 0
        meta["plugmap"] = Table.read(hdulist[5])[0:limit]

    return Spectrum(flux=flux,
                      wcs=fixed_wcs,
                      uncertainty=uncertainty,
                      meta=meta,
                      mask=mask,
                      spectral_axis_index=-1)



import asdf
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import FK5
from astropy.nddata import StdDevUncertainty

from specutils import Spectrum, SpectrumList, SpectralAxis
from specutils.io.asdf.tests.helpers import (
    assert_spectrum_equal, assert_spectrumlist_equal, assert_spectral_axis_equal)


def create_spectrum(xmin, xmax, uncertainty=False, mask=False):
    flux = np.ones(10) * u.Jy
    wavelength = np.linspace(xmin, xmax, 10) * u.nm
    unc = StdDevUncertainty(flux * 0.1) if uncertainty else None
    msk = np.array([0, 1, 1, 0, 1, 0, 1, 1, 0, 1], dtype=np.uint8) if mask else None
    return Spectrum(spectral_axis=wavelength, flux=flux, uncertainty=unc, mask=msk)


@pytest.mark.parametrize('uncertainty', [False, True])
@pytest.mark.parametrize('mask', [False, True])
def test_asdf_spectrum(tmp_path, uncertainty, mask):
    file_path = tmp_path / "test.asdf"
    spectrum = create_spectrum(510, 530, uncertainty=uncertainty, mask=mask)
    with asdf.AsdfFile() as af:
        af["spectrum"] = spectrum
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_spectrum_equal(af["spectrum"], spectrum)


def test_asdf_spectralaxis(tmp_path):
    file_path = tmp_path / "test.asdf"
    wavelengths  = np.arange(510, 530) * u.nm
    spectral_axis = SpectralAxis(wavelengths, bin_specification="edges")

    with asdf.AsdfFile() as af:
        af["spectral_axis"] = spectral_axis
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_spectral_axis_equal(af["spectral_axis"], spectral_axis)


def test_asdf_spectrumlist(tmp_path):
    file_path = tmp_path / "test.asdf"
    spectra = SpectrumList([
        create_spectrum(510, 530),
        create_spectrum(500, 550),
        create_spectrum(0, 10),
        create_spectrum(0.1, 0.5)
    ])
    with asdf.AsdfFile() as af:
        af["spectrum_list"] = spectra
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_spectrumlist_equal(af["spectrum_list"], spectra)


def test_asdf_url_mapper():
    """Make sure specutils ASDF extension url_mapping does not interfere with astropy schemas."""
    with asdf.AsdfFile() as af:
        af.tree = {'frame': FK5()}



"""
Loader for PFS spectrum files.

https://github.com/Subaru-PFS/datamodel/blob/master/datamodel.txt
"""
import os
import re

from astropy.units import Unit
from astropy.nddata import StdDevUncertainty

import numpy as np

from ...spectra import Spectrum
from ..registers import data_loader
from ..parsing_utils import _fits_identify_by_name, read_fileobj_or_hdulist

__all__ = ['identify_pfs_spec', 'pfs_spec_loader']

# This RE matches the file name pattern defined in Subaru-PFS' datamodel.txt :
# "pfsObject-%05d-%s-%3d-%08x-%02d-0x%08x.fits" % (tract, patch, catId, objId,
#                                                  nVisit % 100, pfsVisitHash)
_spec_pattern = re.compile(r'pfsObject-(?P<tract>\d{5})-(?P<patch>.{3})-'
                           r'(?P<catId>\d{3})-(?P<objId>\d{8})-'
                           r'(?P<nVisit>\d{2})-(?P<pfsVisitHash>0x\w{8})'
                           r'\.fits')


def identify_pfs_spec(origin, *args, **kwargs):
    """
    Check whether given file is FITS and name matches `_spec_pattern`.
    """

    return _fits_identify_by_name(origin, *args, pattern=_spec_pattern)


@data_loader(
    label="Subaru-pfsObject", identifier=identify_pfs_spec, extensions=['fits'],
    priority=10,
)
def pfs_spec_loader(file_obj, **kwargs):
    """
    Loader for PFS combined spectrum files.

    Parameters
    ----------
    file_obj : str or file-like
        FITS file name or object (provided from name by Astropy I/O Registry).

    Returns
    -------
    data : Spectrum
        The spectrum that is represented by the data in this table.
    """

    # This will fail for file-like objects without 'name' property like `bz2.BZ2File`,
    # workarund needed (or better yet, a scheme to parse the `meta` items from the header).
    if isinstance(file_obj, str):
        file_name = file_obj
    else:
        file_name = file_obj.name

    m = _spec_pattern.match(os.path.basename(file_name))

    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        meta = {'header': header,
                'tract': m['tract'],
                'patch': m['patch'],
                'catId': m['catId'],
                'objId': m['objId'],
                'nVisit': m['nVisit'],
                'pfsVisitHash': m['pfsVisitHash']}

        # spectrum is in HDU 2
        data = hdulist[2].data['flux']
        unit = Unit('nJy')

        error = hdulist[2].data['fluxVariance']
        uncertainty = StdDevUncertainty(np.sqrt(error))

        wave = hdulist[2].data['lambda']
        wave_unit = Unit('nm')

        mask = hdulist[2].data['mask'] != 0

    return Spectrum(flux=data * unit,
                      spectral_axis=wave * wave_unit,
                      uncertainty=uncertainty,
                      meta=meta,
                      mask=mask)



import warnings
from copy import deepcopy

import numpy as np
from astropy import units as u
from astropy.coordinates import SpectralCoord
from astropy.utils.decorators import lazyproperty
from astropy.utils.decorators import deprecated
from astropy.nddata import NDUncertainty, NDIOMixin, NDArithmeticMixin
from gwcs.wcs import WCS as GWCS

from .spectral_axis import SpectralAxis
from .spectrum_mixin import OneDSpectrumMixin
from .spectral_region import SpectralRegion
from ..utils.wcs_utils import gwcs_from_array

from ndcube import NDCube

__all__ = ['Spectrum1D', 'Spectrum']


class Spectrum(OneDSpectrumMixin, NDCube, NDIOMixin, NDArithmeticMixin):
    """
    Spectrum container for N-dimensional data with one spectral axis.

    `Spectrum` can contain "vector spectra" by having the
    ``flux`` have a shape with dimension greater than 1.  One dimension of
    ``flux`` must match the length of the ``spectral_axis`` if ``spectral_axis``
    is provided.

    For multidimensional spectra that are all the same shape but have different
    spectral axes, use a :class:`~specutils.SpectrumCollection`.  For a
    collection of spectra that have different shapes, use
    :class:`~specutils.SpectrumList`. For more on this topic, see
    :ref:`specutils-representation-overview`.

    Parameters
    ----------
    flux : `~astropy.units.Quantity`
        The flux data for this spectrum. This can be a simple `~astropy.units.Quantity`,
        or an existing `~Spectrum` or `~ndcube.NDCube` object.
    spectral_axis : `~astropy.units.Quantity` or `~specutils.SpectralAxis`
        Dispersion information with the same shape as the last (or only)
        dimension of flux, or one greater than the last dimension of flux
        if specifying bin edges.
    spectral_axis_index : integer, optional
        If it is ambiguous which axis is the spectral axis (e.g., if there are multiple
        axes in the flux array with the same length as the input spectral_axis),
        this argument is used to specify which is the spectral axis.
    move_spectral_axis : int, str, optional
        Force the spectral axis to be either the last axis (the default behavior prior
        to version 2.0) by setting this argument to 'last' or -1, or the first axis by
        setting this argument to 'first' or 0.
        This will do a simple ``swapaxis`` between the relevant axis and original
        spectral axis. If None, the spectral axis is left wherever it is in the input.
    wcs : `~astropy.wcs.WCS` or `~gwcs.wcs.WCS`
        WCS information object that either has a spectral component or is
        only spectral.
    velocity_convention : {"relativistic", "optical", "radio"}
        Convention used for velocity conversions.
    rest_value : `~astropy.units.Quantity`
        Any quantity supported by the standard spectral equivalencies
        (wavelength, energy, frequency, wave number). Describes the rest value
        of the spectral axis for use with velocity conversions.
    redshift
        See `redshift` for more information.
    radial_velocity
        See `radial_velocity` for more information.
    bin_specification : str
        Either "edges" or "centers" to indicate whether the `spectral_axis`
        values represent edges of the wavelength bin, or centers of the bin.
    uncertainty : `~astropy.nddata.NDUncertainty`
        Contains uncertainty information along with propagation rules for
        spectrum arithmetic. Can take a unit, but if none is given, will use
        the unit defined in the flux.
    mask : `~numpy.ndarray`-like
        Array where values in the flux to be masked are those that
        ``astype(bool)`` converts to True. (For example, integer arrays are not
        masked where they are 0, and masked for any other value.)
    meta : dict
        Arbitrary container for any user-specific information to be carried
        around with the spectrum container object.
    """
    def __init__(self, flux=None, spectral_axis=None, spectral_axis_index=None,
                 wcs=None, velocity_convention=None, rest_value=None,
                 redshift=None, radial_velocity=None, bin_specification=None,
                 move_spectral_axis=None, **kwargs):

        # If the flux (data) argument is already a Spectrum (as it would
        # be for internal arithmetic operations), avoid setup entirely.
        if isinstance(flux, Spectrum):
            self._spectral_axis_index = flux.spectral_axis_index
            self._spectral_axis = flux.spectral_axis
            super().__init__(flux)
            return

        self._spectral_axis_index = spectral_axis_index
        # Might as well handle this right away
        if spectral_axis_index is None and flux is not None:
            if flux.ndim == 1:
                self._spectral_axis_index = 0
        elif flux is None:
            self._spectral_axis_index = 0

        # Check for pre-defined entries in the kwargs dictionary.
        unknown_kwargs = set(kwargs).difference(
            {'data', 'unit', 'uncertainty', 'meta', 'mask', 'copy',
             'extra_coords'})

        if len(unknown_kwargs) > 0:
            raise ValueError("Initializer contains unknown arguments(s): {}."
                             "".format(', '.join(map(str, unknown_kwargs))))

        # Handle initializing from NDCube objects
        elif isinstance(flux, NDCube):
            if flux.unit is None:
                raise ValueError("Input NDCube missing unit parameter")

            # Change the flux array from bare ndarray to a Quantity
            q_flux = flux.data << u.Unit(flux.unit)

            self.__init__(flux=q_flux, wcs=flux.wcs, mask=flux.mask,
                          uncertainty=flux.uncertainty)
            return

        # If the mask kwarg is not passed to the constructor, but the flux array
        # contains NaNs, add the NaN locations to the mask.
        if "mask" not in kwargs and flux is not None:
            nan_mask = np.isnan(flux)
            if nan_mask.any():
                if hasattr(self, "mask"):
                    kwargs["mask"] = np.logical_or(nan_mask, self.mask)
                else:
                    kwargs["mask"] = nan_mask.copy()
            del nan_mask

        # Ensure that the flux argument is an astropy quantity
        if flux is not None:
            if not isinstance(flux, u.Quantity):
                raise ValueError("Flux must be a `Quantity` object.")
            elif flux.isscalar:
                flux = u.Quantity([flux])

        # Ensure that only one or neither of these parameters is set
        if redshift is not None and radial_velocity is not None:
            raise ValueError("Cannot set both radial_velocity and redshift at "
                             "the same time.")

        # In cases of slicing, new objects will be initialized with `data`
        # instead of ``flux``. Ensure we grab the `data` argument.
        if flux is None and 'data' in kwargs:
            flux = kwargs.pop('data')

        # Ensure that the unit information codified in the quantity object is
        # the One True Unit.
        kwargs.setdefault('unit', flux.unit if isinstance(flux, u.Quantity)
                                            else kwargs.get('unit'))

        # In the case where the arithmetic operation is being performed with
        # a single float, int, or array object, just go ahead and ignore wcs
        # requirements
        if np.ndim(flux) == 0 and spectral_axis is None and wcs is None:
            super(Spectrum, self).__init__(data=flux, wcs=wcs, **kwargs)
            return

        if rest_value is None:
            if hasattr(wcs, 'rest_frequency') and wcs.rest_frequency != 0:
                rest_value = wcs.rest_frequency * u.Hz
            elif hasattr(wcs, 'rest_wavelength') and wcs.rest_wavelength != 0:
                rest_value = wcs.rest_wavelength * u.AA
            elif hasattr(wcs, 'wcs') and hasattr(wcs.wcs, 'restfrq') and wcs.wcs.restfrq > 0:
                rest_value = wcs.wcs.restfrq * u.Hz
            elif hasattr(wcs, 'wcs') and hasattr(wcs.wcs, 'restwav') and wcs.wcs.restwav > 0:
                rest_value = wcs.wcs.restwav * u.m
            else:
                rest_value = None
        else:
            if not isinstance(rest_value, u.Quantity):
                warnings.warn("No unit information provided with rest value. "
                              f"Assuming units of spectral axis ('{spectral_axis.unit}').")
                rest_value = u.Quantity(rest_value, spectral_axis.unit)
            elif not rest_value.unit.is_equivalent(u.AA, equivalencies=u.spectral()):
                raise u.UnitsError("Rest value must be "
                                   "energy/wavelength/frequency equivalent.")

        # If flux and spectral axis are both specified, check that their lengths
        # match or are off by one (implying the spectral axis stores bin edges).
        # If we can't determine which flux axis corresponds to the spectral axis
        # we raise an error.
        if flux is not None and spectral_axis is not None:
            if spectral_axis_index is None:
                if flux.ndim == 1:
                    self._spectral_axis_index = 0
                else:
                    matching_axes = []
                    if bin_specification == "centers":
                        add_elements = [0,]
                    elif bin_specification == "edges":
                        add_elements = [1,]
                    elif bin_specification is None:
                        add_elements = [0,1]
                    for i in range(flux.ndim):
                        for add_element in add_elements:
                            if spectral_axis.shape[0] == flux.shape[i] + add_element:
                                matching_axes.append(i)

                    if len(matching_axes) == 1:
                        self._spectral_axis_index = matching_axes[0]
                    else:
                        raise ValueError("Unable to determine which flux axis corresponds to "
                                         "the spectral axis. Please specify spectral_axis_index"
                                         " or provide a spectral_axis matching a flux axis.")

            # Make sure the length of the spectral axis matches the appropriate flux axis
            if spectral_axis.shape[0] == flux.shape[self.spectral_axis_index]:
                if bin_specification == "edges":
                    raise ValueError("A spectral axis input as bin edges must "
                                     "have length one greater than the flux axis")
                bin_specification = "centers"
            elif spectral_axis.shape[0] == flux.shape[self.spectral_axis_index]+1:
                if bin_specification == "centers":
                    raise ValueError("A spectral axis input as bin centers "
                        "must be the same length as the flux axis")
                bin_specification = "edges"
            else:
                raise ValueError(
                    f"Spectral axis length ({spectral_axis.shape[0]}) must be the "
                    "same size or one greater (if specifying bin edges) than that "
                    f"of the corresponding flux axis ({flux.shape[self.spectral_axis_index]})")

        # If a WCS is provided, determine which axis is the spectral axis
        if wcs is not None:
            naxis = None
            if hasattr(wcs, "naxis"):
                naxis = wcs.naxis
            # GWCS doesn't have naxis
            elif hasattr(wcs, "world_n_dim"):
                naxis = wcs.world_n_dim

            if naxis is not None and naxis > 1:
                temp_axes = []
                phys_axes = wcs.world_axis_physical_types
                if self._spectral_axis_index is None:
                    for i in range(len(phys_axes)):
                        if phys_axes[i] is None:
                            continue
                        if phys_axes[i][0:2] == "em" or phys_axes[i][0:5] == "spect":
                            temp_axes.append(i)
                    if len(temp_axes) != 1:
                        raise ValueError("Input WCS must have exactly one axis with "
                                        "spectral units, found {}".format(len(temp_axes)))
                    else:
                        # Due to FITS conventions, the WCS axes are listed in opposite
                        # order compared to the data array.
                        self._spectral_axis_index = len(flux.shape)-temp_axes[0]-1

                if move_spectral_axis is not None:
                    if isinstance(wcs, GWCS):
                        raise ValueError("move_spectral_axis cannot be used with GWCS")
                    if isinstance(move_spectral_axis, str):
                        if move_spectral_axis.lower() == 'first':
                            move_to_index = 0
                        elif move_spectral_axis.lower() == 'last':
                            move_to_index = wcs.naxis - 1
                        else:
                            raise ValueError("move_spectral_axis must be either 'first' or 'last'")
                    elif isinstance(move_spectral_axis, int):
                        move_to_index = move_spectral_axis
                    else:
                        raise ValueError("move_spectral_axis must be an integer or 'first'/'last'")

                    if move_to_index != self._spectral_axis_index:
                        wcs = wcs.swapaxes(self._spectral_axis_index, move_to_index)
                        if flux is not None:
                            flux = np.swapaxes(flux, self._spectral_axis_index, move_to_index)
                        if "mask" in kwargs:
                            if kwargs["mask"] is not None:
                                kwargs["mask"] = np.swapaxes(kwargs["mask"],
                                                    self._spectral_axis_index, move_to_index)
                        if "uncertainty" in kwargs:
                            if kwargs["uncertainty"] is not None:
                                if isinstance(kwargs["uncertainty"], NDUncertainty):
                                    # Account for Astropy uncertainty types
                                    temp_unc = np.swapaxes(kwargs["uncertainty"].array,
                                                           self._spectral_axis_index, move_to_index)
                                    if kwargs["uncertainty"].unit is not None:
                                        temp_unc = temp_unc * u.Unit(kwargs["uncertainty"].unit)
                                    kwargs["uncertainty"] = type(kwargs["uncertainty"])(temp_unc)
                                else:
                                    kwargs["uncertainty"] = np.swapaxes(kwargs["uncertainty"],
                                                            self._spectral_axis_index, move_to_index)

                        self._spectral_axis_index = move_to_index

            else:
                if flux is not None and flux.ndim == 1:
                    self._spectral_axis_index = 0
                else:
                    if self.spectral_axis_index is None:
                        raise ValueError("WCS is 1D but flux is multi-dimensional. Please"
                                         " specify spectral_axis_index.")

        elif move_spectral_axis is not None:
            raise ValueError("Unable to use `move_spectral_axis` without a multi-dimensional WCS")

        # Attempt to parse the spectral axis. If none is given, try instead to
        # parse a given wcs. This is put into a GWCS object to
        # then be used behind-the-scenes for all specutils operations.
        if spectral_axis is not None:
            # Ensure that the spectral axis is an astropy Quantity
            if not isinstance(spectral_axis, u.Quantity):
                raise ValueError("Spectral axis must be a `Quantity` or "
                                 "`SpectralAxis` object.")

            # If spectral axis is provided as an astropy Quantity, convert it
            # to a specutils SpectralAxis object.
            if not isinstance(spectral_axis, SpectralAxis):
                self._spectral_axis = SpectralAxis(
                    spectral_axis, redshift=redshift,
                    radial_velocity=radial_velocity, doppler_rest=rest_value,
                    doppler_convention=velocity_convention,
                    bin_specification=bin_specification)
            # If a SpectralAxis object is provided, we assume it doesn't need
            # information from other keywords added
            else:
                for a in [radial_velocity, redshift]:
                    if a is not None:
                        raise ValueError("Cannot separately set redshift or "
                                         "radial_velocity if a SpectralAxis "
                                         "object is input to spectral_axis")

                self._spectral_axis = spectral_axis

            if wcs is None:
                wcs = gwcs_from_array(self._spectral_axis,
                                      flux.shape,
                                      spectral_axis_index=self.spectral_axis_index
                                      )

        elif wcs is None:
            # If no spectral axis or wcs information is provided, initialize
            # with an empty gwcs based on the flux.
            if self.spectral_axis_index is None:
                if flux.ndim == 1:
                    self._spectral_axis_index = 0
                else:
                    raise ValueError("Must specify spectral_axis_index if no WCS or spectral"
                                     " axis is input.")
            size = flux.shape[self.spectral_axis_index] if not flux.isscalar else 1
            wcs = gwcs_from_array(np.arange(size) * u.Unit(""),
                                  flux.shape,
                                  spectral_axis_index=self.spectral_axis_index
                                  )

        super().__init__(
            data=flux.value if isinstance(flux, u.Quantity) else flux,
            wcs=wcs, **kwargs)

        # If no spectral_axis was provided, create a SpectralCoord based on
        # the WCS
        if spectral_axis is None:
            # If the WCS doesn't have a spectral attribute, we assume it's the
            # dummy GWCS we created or a solely spectral WCS
            if hasattr(self.wcs, "spectral"):
                # Handle generated 1D WCS that aren't set to spectral
                if not self.wcs.is_spectral and self.wcs.naxis == 1:
                    spec_axis = self.wcs.pixel_to_world(
                                    np.arange(self.flux.shape[self.spectral_axis_index]))
                else:
                    spec_axis = self.wcs.spectral.pixel_to_world(
                                    np.arange(self.flux.shape[self.spectral_axis_index]))
            else:
                # We now keep the entire GWCS, including spatial information, so we need to include
                # all axes in the pixel_to_world call. Note that this assumes/requires that the
                # dispersion is the same at all spatial locations.
                wcs_args = []
                for i in range(len(self.flux.shape)):
                    wcs_args.append(np.zeros(self.flux.shape[self.spectral_axis_index]))
                # Replace with arange for the spectral axis
                wcs_args[self.spectral_axis_index] = np.arange(self.flux.shape[self.spectral_axis_index])
                wcs_args.reverse()
                temp_coords = self.wcs.pixel_to_world(*wcs_args)
                # If there are spatial axes, temp_coords will have a SkyCoord and a SpectralCoord
                if isinstance(temp_coords, list):
                    for coords in temp_coords:
                        if isinstance(coords, SpectralCoord):
                            spec_axis = coords
                            break
                    else:
                        # WCS axis ordering is reverse of numpy
                        spec_axis = temp_coords[len(temp_coords) - self.spectral_axis_index - 1]
                else:
                    spec_axis = temp_coords

            try:
                if spec_axis.unit.is_equivalent(u.one):
                    spec_axis = spec_axis * u.pixel
            except AttributeError:
                raise AttributeError(f"spec_axis does not have unit: "
                                     f"{type(spec_axis)} {spec_axis}")

            self._spectral_axis = SpectralAxis(
                spec_axis,
                redshift=redshift, radial_velocity=radial_velocity,
                doppler_rest=rest_value,
                doppler_convention=velocity_convention)

        # make sure that spectral axis is strictly increasing or decreasing,
        # raise an error if not.
        if not self._check_strictly_increasing_decreasing():
            raise ValueError('Spectral axis must be strictly increasing or decreasing.')

        if hasattr(self, 'uncertainty') and self.uncertainty is not None:
            if not flux.shape == self.uncertainty.array.shape:
                raise ValueError(
                    "Flux axis ({}) and uncertainty ({}) shapes must be the "
                    "same.".format(flux.shape, self.uncertainty.array.shape))

    def __getitem__(self, item):
        """
        Override the class indexer. We do this here because there are two cases
        for slicing on a ``Spectrum``:

            1.) When the flux is one dimensional, indexing represents a single
                flux value at a particular spectral axis bin, and returns a new
                ``Spectrum`` where all attributes are sliced.
            2.) When flux is multi-dimensional (i.e. several fluxes over the
                same spectral axis), indexing returns a new ``Spectrum`` with
                the sliced flux range and a deep copy of all other attributes.

        The first case is handled by the parent class, while the second is
        handled here.
        """
        new_spectral_axis_index = self.spectral_axis_index

        if self.flux.ndim > 1 or (isinstance(item, tuple) and item[0] is Ellipsis):
            if isinstance(item, tuple):
                if len(item) == self.flux.ndim or item[0] == Ellipsis:
                    spec_item = item[self.spectral_axis_index]
                    if not isinstance(spec_item, slice):
                        if isinstance(item, u.Quantity):
                            raise ValueError("Indexing on single spectral axis "
                                             "values is not currently allowed, "
                                             "please use a slice.")
                        spec_item = slice(spec_item, spec_item+1, None)
                        # We have to hack around updating this tuple entry
                        item = list(item)
                        item[self.spectral_axis_index] = spec_item
                        item = tuple(item)
                else:
                    # Slicing on less than the full number of axes means we want
                    # to keep the whole spectral axis
                    spec_item = slice(None, None, None)
                    # If any slices are single integers, need to decrement the spectral axis index

                for i in range(len(item)-1):
                    # Decrement spectral_axis_index for each single element slice
                    if isinstance(item[i], int):
                        new_spectral_axis_index -= 1

            elif isinstance(item, slice) and (isinstance(item.start, u.Quantity) or
                    isinstance(item.stop, u.Quantity)):
                # We only allow slicing with world coordinates along the spectral
                # axis for now
                for attr in ("start", "stop"):
                    if getattr(item, attr) is None:
                        continue
                    if not getattr(item, attr).unit.is_equivalent(u.AA,
                            equivalencies=u.spectral()):
                        raise ValueError("Slicing with world coordinates is only"
                                         " enabled for spectral coordinates.")
                        break
                spec_item = item
            else:
                # Slicing with a single integer or slice uses the leading axis,
                # so we keep the whole spectral axis, which is last. Need to decrement
                # the spectral axis index by one
                if isinstance(item, int):
                    new_spectral_axis_index -= 1
                spec_item = slice(None, None, None)

            if (isinstance(spec_item.start, u.Quantity) or
                    isinstance(spec_item.stop, u.Quantity)):
                temp_spec = self._spectral_slice(spec_item)
                if spec_item is item:
                    return temp_spec
                else:
                    # Drop the spectral axis slice and perform only the spatial part
                    return temp_spec[item[:-1]]

            if "original_wcs" not in self.meta:
                new_meta = deepcopy(self.meta)
                new_meta["original_wcs"] = deepcopy(self.wcs)
            else:
                new_meta = deepcopy(self.meta)

            return self._copy(
                flux=self.flux[item],
                spectral_axis=self.spectral_axis[spec_item],
                uncertainty=self.uncertainty[item]
                if self.uncertainty is not None else None,
                mask=self.mask[item] if self.mask is not None else None,
                meta=new_meta, wcs=None, spectral_axis_index=new_spectral_axis_index)

        if not isinstance(item, slice):
            if isinstance(item, u.Quantity):
                raise ValueError("Indexing on a single spectral axis value is not"
                                 " currently allowed, please use a slice.")
            # Handle tuple slice as input by NDCube crop method
            elif isinstance(item, tuple):
                if len(item) == 1 and isinstance(item[0], slice):
                    item = item[0]
                else:
                    raise ValueError(f"Unclear how to slice with tuple {item}")
            else:
                item = slice(item, item + 1, None)
        elif (isinstance(item.start, u.Quantity) or isinstance(item.stop, u.Quantity)):
            return self._spectral_slice(item)

        # Work around error in SpectralCoord creation in super().__getitem__ for
        # spectral axis in pixels.
        if self.spectral_axis.unit == u.pix:
            if "original_wcs" not in self.meta:
                new_meta = deepcopy(self.meta)
                new_meta["original_wcs"] = deepcopy(self.wcs)
            return self._copy(
                flux=self.flux[item],
                spectral_axis=self.spectral_axis[item],
                uncertainty=self.uncertainty[item]
                if self.uncertainty is not None else None,
                mask=self.mask[item] if self.mask is not None else None,
                meta=new_meta, wcs=None, spectral_axis_index=new_spectral_axis_index)

        tmp_spec = super().__getitem__(item)

        # TODO: this is a workaround until we figure out how to deal with non-
        #  strictly ascending spectral axes. Currently, the wcs is created from
        #  a spectral axis array by converting to a length physical type. On
        #  a regular slicing operation, the wcs is handed back to the
        #  initializer and a new spectral axis is created. This would then also
        #  be in length units, which may not be the units used initially. So,
        #  we create a new ``Spectrum`` that includes the sliced spectral
        #  axis. This means that a new wcs object will be created with the
        #  appropriate unit translation handling.
        if "original_wcs" not in self.meta:
            new_meta = deepcopy(self.meta)
            new_meta["original_wcs"] = deepcopy(self.wcs)
        else:
            new_meta = deepcopy(self.meta)

        return tmp_spec._copy(spectral_axis=self.spectral_axis[item], wcs=None,
                              meta=new_meta)

    def _copy(self, **kwargs):
        """
        Perform deep copy operations on each attribute of the ``Spectrum``
        object.
        """
        alt_kwargs = dict(
            flux=deepcopy(self.flux),
            spectral_axis=deepcopy(self.spectral_axis),
            uncertainty=deepcopy(self.uncertainty),
            wcs=deepcopy(self.wcs),
            mask=deepcopy(self.mask),
            meta=deepcopy(self.meta),
            unit=deepcopy(self.unit),
            velocity_convention=deepcopy(self.velocity_convention),
            rest_value=deepcopy(self.rest_value),
            spectral_axis_index=deepcopy(self.spectral_axis_index))

        alt_kwargs.update(kwargs)

        return self.__class__(**alt_kwargs)

    def _spectral_slice(self, item):
        """
        Perform a region extraction given a slice on the spectral axis.
        """
        from ..manipulation import extract_region
        if item.start is None:
            start = self.spectral_axis[0]
        else:
            start = item.start
        if item.stop is None:
            stop = self.spectral_axis[-1]
        else:
            # Force the upper bound to be open, as in normal python array slicing
            exact_match = np.where(self.spectral_axis == item.stop)
            if len(exact_match[0]) == 1:
                stop_index = exact_match[0][0] - 1
                stop = self.spectral_axis[stop_index]
            else:
                stop = item.stop
        reg = SpectralRegion(start, stop)
        return extract_region(self, reg)

    def collapse(self, method, axis=None):
        """
        Collapse the flux array given a method. Will collapse either to a single
        value (default), over a specified numerical axis or axes if specified, or
        over the spectral or non-spectral axes if ``physical_type`` is specified.

        If the collapse leaves the spectral axis unchanged, a `~specutils.Spectrum`
        will be returned. Otherwise an `~astropy.units.Quantity` array will be
        returned.

        Note that these calculations are not currently uncertainty-aware, but do
        respect masks.

        Parameters
        ----------

        method : str, function
            The method by which the flux will be collapsed. String options are
            'mean', 'min', 'max', 'sum', and 'median'. Also accepts a function
            as input, which must take an `astropy.units.Quantity` array as input
            and accept an 'axis' argument.
        axis : int, tuple, str, optional
            The axis or axes over which to collapse the flux array. May also be
            a string, either 'spectral' to collapse over the spectral axis, or
            'spatial' to collapse over all other axes.

        Returns
        -------
        :class:`~specutils.Spectrum` or :class:`~astropy.units.Quantity`

        """
        collapse_funcs = {"mean": np.nanmean, "max": np.nanmax, "min": np.nanmin,
                         "median": np.nanmedian, "sum": np.nansum}

        if isinstance(axis, str):
            if axis == 'spectral':
                axis = self.spectral_axis_index
            elif axis == 'spatial':
                # generate tuple if needed for multiple spatial axes
                axis = tuple([x for x in range(len(self.flux.shape)) if
                              x != self.spectral_axis_index])
            else:
                raise ValueError("String axis input must be 'spatial' or 'spectral'")

        # Set masked locations to NaN for the calculation, since the `where` argument
        # does not seem to work consistently in the numpy functions.
        flux_to_collapse = self.flux.copy()
        if self.mask is not None:
            flux_to_collapse[np.where(self.mask != 0)] = np.nan

        # Leave open the possibility of the user providing their own method
        if callable(method):
            collapsed_flux = method(flux_to_collapse, axis=axis)
        else:
            collapsed_flux = collapse_funcs[method](flux_to_collapse, axis=axis)

        # Return a Spectrum if we collapsed over the spectral axis, a Quantity if not
        if axis in (self.spectral_axis_index, None):
            return collapsed_flux
        elif isinstance(axis, tuple) and self.spectral_axis_index in axis:
            return collapsed_flux
        else:
            # Pass the spectral axis rather than WCS in this case, so we don't have to
            # figure out which part of a multidimensional WCS is the spectral part.
            return Spectrum(collapsed_flux, spectral_axis=self.spectral_axis)

    def mean(self, **kwargs):
        return self.collapse("mean", **kwargs)

    def max(self, **kwargs):
        return self.collapse("max", **kwargs)

    def min(self, **kwargs):
        return self.collapse("min", **kwargs)

    def median(self, **kwargs):
        return self.collapse("median", **kwargs)

    def sum(self, **kwargs):
        return self.collapse("sum", **kwargs)

    @NDCube.mask.setter
    def mask(self, value):
        # Impose stricter checks than the base NDData mask setter
        if value is not None:
            value = np.array(value)
            if not self.data.shape == value.shape:
                raise ValueError(
                    "Flux axis ({}) and mask ({}) shapes must be the "
                    "same.".format(self.data.shape, value.shape))
        self._mask = value

    @property
    def frequency(self):
        """
        The `spectral_axis` as a `~astropy.units.Quantity` in units of GHz
        """
        return self.spectral_axis.to(u.GHz, u.spectral())

    @property
    def wavelength(self):
        """
        The `spectral_axis` as a `~astropy.units.Quantity` in units of Angstroms
        """
        return self.spectral_axis.to(u.AA, u.spectral())

    @property
    def energy(self):
        """
        The energy of the spectral axis as a `~astropy.units.Quantity` in units
        of eV.
        """
        return self.spectral_axis.to(u.eV, u.spectral())

    @property
    def photon_flux(self):
        """
        The flux density of photons as a `~astropy.units.Quantity`, in units of
        photons per cm^2 per second per spectral_axis unit
        """
        flux_in_spectral_axis_units = self.flux.to(
            u.W * u.cm**-2 * self.spectral_axis.unit**-1,
            u.spectral_density(self.spectral_axis))
        photon_flux_density = flux_in_spectral_axis_units / (self.energy / u.photon)

        return photon_flux_density.to(u.photon * u.cm**-2 * u.s**-1 *
                                      self.spectral_axis.unit**-1)

    @lazyproperty
    def bin_edges(self):
        return self.spectral_axis.bin_edges

    @property
    def shape(self):
        return self.flux.shape

    @property
    def spectral_axis_direction(self):
        return self._spectral_axis_direction

    @property
    def redshift(self):
        """
        The redshift(s) of the objects represented by this spectrum.  May be
        scalar (if this spectrum's ``flux`` is 1D) or vector.  Note that
        the concept of "redshift of a spectrum" can be ambiguous, so the
        interpretation is set to some extent by either the user, or operations
        (like template fitting) that set this attribute when they are run on
        a spectrum.
        """
        return self.spectral_axis.redshift

    @property
    def radial_velocity(self):
        """
        The radial velocity(s) of the objects represented by this spectrum.  May
        be scalar (if this spectrum's ``flux`` is 1D) or vector.  Note that
        the concept of "RV of a spectrum" can be ambiguous, so the
        interpretation is set to some extent by either the user, or operations
        (like template fitting) that set this attribute when they are run on
        a spectrum.
        """
        return self.spectral_axis.radial_velocity

    def set_redshift_to(self, redshift):
        """
        This sets the redshift of the spectrum to be `redshift` *without*
        changing the values of the `spectral_axis`.

        If you want to shift the `spectral_axis` based on this value, use
        `shift_spectrum_to`.
        """
        new_spec_coord = self.spectral_axis.replicate(redshift=redshift)
        self._spectral_axis = new_spec_coord

    def set_radial_velocity_to(self, radial_velocity):
        """
        This sets the radial velocity of the spectrum to be `radial_velocity`
        *without* changing the values of the `spectral_axis`.

        If you want to shift the `spectral_axis` based on this value, use
        `shift_spectrum_to`.
        """
        new_spec_coord = self.spectral_axis.replicate(
            radial_velocity=radial_velocity
        )
        self._spectral_axis = new_spec_coord

    def shift_spectrum_to(self, *, redshift=None, radial_velocity=None):
        """
        This shifts in-place the values of the `spectral_axis`, given either a
        redshift or radial velocity.

        If you do *not* want to change the `spectral_axis`, use
        `set_redshift_to` or `set_radial_velocity_to`.
        """
        if redshift is not None and radial_velocity is not None:
            raise ValueError(
                "Only one of redshift or radial_velocity can be used."
            )
        if redshift is not None:
            new_spec_coord = self.spectral_axis.with_radial_velocity_shift(
                -self.spectral_axis.radial_velocity
            ).with_radial_velocity_shift(redshift)
            self._spectral_axis = new_spec_coord
        elif radial_velocity is not None:
            if radial_velocity is not None:
                if not radial_velocity.unit.is_equivalent(u.km/u.s):
                    raise u.UnitsError("Radial velocity must be a velocity.")

            new_spectral_axis = self.spectral_axis.with_radial_velocity_shift(
                -self.spectral_axis.radial_velocity
            ).with_radial_velocity_shift(radial_velocity)
            self._spectral_axis = new_spectral_axis
        else:
            raise ValueError("One of redshift or radial_velocity must be set.")

    def with_spectral_axis_last(self):
        """
        Convenience method to return a new copy of the Spectrum with the spectral axis last.
        """
        return Spectrum(flux=self.flux, wcs=self.wcs,
                          mask=self.mask, uncertainty=self.uncertainty,
                          redshift=self.redshift, move_spectral_axis="last")

    @redshift.setter
    @deprecated('1.8.0', alternative='set_redshift_to or shift_spectrum_to')
    def redshift(self, val):
        self.shift_spectrum_to(redshift=val)

    @radial_velocity.setter
    @deprecated('1.8.0', alternative='set_radial_velocity_to or shift_spectrum_to')
    def radial_velocity(self, val):
        self.shift_spectrum_to(radial_velocity=val)

    def _return_with_redshift(self, result):
        result.shift_spectrum_to(redshift=self.redshift)
        return result

    def _other_as_correct_class(self, other, force_quantity=False):
        # NDArithmetic mixin will try to turn other into a Spectrum, which will fail
        # sometimes because of not specifiying the spectral axis index
        if isinstance(other, Spectrum):
            # Take this opportunity to check if the spectral axes match
            if not np.all(other.spectral_axis == self.spectral_axis):
                raise ValueError("Spectral axis of both operands must match")
        else:
            if not isinstance(other, u.Quantity) and force_quantity:
                other = other * self.unit

        if isinstance(other, u.Quantity) and other.shape == self.shape:
            return Spectrum(flux=other, spectral_axis=self.spectral_axis,
                            spectral_axis_index=self.spectral_axis_index)

        return other

    def __add__(self, other):
        other = self._other_as_correct_class(other, force_quantity=True)
        if isinstance(other, (Spectrum)):
            return self._return_with_redshift(self.add(other))
        else:
            new_flux = self.flux + other
            return self._return_with_redshift(Spectrum(new_flux, wcs=self.wcs, meta=self.meta,
                                                       uncertainty=self.uncertainty))

    def __sub__(self, other):
        other = self._other_as_correct_class(other, force_quantity=True)
        if isinstance(other, (Spectrum)):
            return self._return_with_redshift(self.subtract(other))
        else:
            new_flux = self.flux - other
            return self._return_with_redshift(Spectrum(new_flux, wcs=self.wcs, meta=self.meta,
                                                       uncertainty=self.uncertainty))

    def __mul__(self, other):
        other = self._other_as_correct_class(other)
        if isinstance(other, (Spectrum)):
            return self._return_with_redshift(self.multiply(other))
        else:
            new_flux = self.flux * other
            if self.uncertainty is None:
                new_uncertainty = None
            else:
                new_uncertainty = deepcopy(self.uncertainty)
                new_uncertainty.array *= other
            return self._return_with_redshift(Spectrum(new_flux, wcs=self.wcs,
                                                       meta=self.meta,
                                                       uncertainty=new_uncertainty))

    def __div__(self, other):
        other = self._other_as_correct_class(other)
        if isinstance(other, (Spectrum)):
            return self._return_with_redshift(self.divide(other))
        else:
            new_flux = self.flux / other
            if self.uncertainty is None:
                new_uncertainty = None
            else:
                new_uncertainty = deepcopy(self.uncertainty)
                new_uncertainty.array /= other
            return self._return_with_redshift(Spectrum(new_flux, wcs=self.wcs,
                                                       meta=self.meta,
                                                       uncertainty=self.uncertainty/other))

    def __truediv__(self, other):
        if not isinstance(other, Spectrum):
            other = self._other_as_correct_class(other)

        return self._return_with_redshift(self.divide(other))

    __radd__ = __add__

    __rmul__ = __mul__

    def __rsub__(self, other):
        return -1 * (self - other)

    def _format_array_summary(self, label, array):
        array_str = np.array2string(array, threshold=8, prefix=label)
        if len(array) >= 1:
            mean = np.nanmean(array)
            s = f"{label}{array_str} {array.unit},  mean={mean:.5f}"
            return s
        else:
            return "{:17} [ ],  mean= n/a".format(label+':')

    def __str__(self):
        result = "Spectrum "
        result += "(length={})\n".format(len(self.spectral_axis))

        # Add Flux information
        result += self._format_array_summary('Flux=', self.flux) + '\n'

        # Add information about spectral axis
        result += self._format_array_summary('Spectral Axis=', self.spectral_axis)

        # Add information about uncertainties if available
        if self.uncertainty:
            result += (f'\nUncertainty={type(self.uncertainty).__name__} '
                       f'({np.array2string(self.uncertainty.array, threshold=8)}'
                       f' {self.uncertainty.unit})')

        return result

    def __repr__(self):
        flux_str  = "flux="
        if (self.flux.ndim == 1 and self.flux.size <= 10) or self.flux.size <= 20:
            flux_str += repr(self.flux)
        else:
            flux_summary = f"{self.flux.value.flat[0]} ... {self.flux.value.flat[-1]}"
            flux_str = flux_str + "[" * self.flux.ndim + flux_summary + "]" * self.flux.ndim
            flux_str += f" {self.flux.unit}"

        flux_str += f" (shape={self.flux.shape}, mean={np.nanmean(self.flux):.5f}); "
        # Sometimes this errors if an error occurs during initialization
        if hasattr(self, "_spectral_axis"):
            spectral_axis_str = (repr(self.spectral_axis).split("[")[0] +
                                    np.array2string(self.spectral_axis, threshold=8) +
                                    f" {self.spectral_axis.unit}>")
            spectral_axis_str = f"spectral_axis={spectral_axis_str} (length={len(self.spectral_axis)})"
            inner_str = (flux_str + spectral_axis_str)
        else:
            inner_str = flux_str

        if self.uncertainty is not None:
            inner_str += f"; uncertainty={self.uncertainty.__class__.__name__}"

        result = "<Spectrum({})>".format(inner_str)

        return result


@deprecated(since="2.0", alternative="Spectrum")
class Spectrum1D(Spectrum):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import astropy.units as u
import numpy as np
import pytest
from astropy import time
from astropy.coordinates import (SkyCoord, EarthLocation, ICRS, Galactic,
                                 SpectralCoord, FK5)
from astropy.utils.exceptions import AstropyDeprecationWarning

from ..spectra.spectral_axis import SpectralAxis
from ..spectra.spectrum import Spectrum

from astropy.tests.helper import assert_quantity_allclose


def get_greenwich_earthlocation():
    """
    A helper function to get an EarthLocation for greenwich (without trying to
    do a download)
    """
    site_registry = EarthLocation._get_site_registry(force_builtin=True)
    return site_registry.get('greenwich')


def test_create_spectral_axis():

    site = get_greenwich_earthlocation()
    obstime = time.Time('2018-12-13 9:00')

    observer_gcrs = site.get_gcrs(obstime)

    wavelengths = np.linspace(500, 2500, 1001) * u.AA
    spectral_axis = SpectralAxis(wavelengths, observer=observer_gcrs)

    assert isinstance(spectral_axis, u.Quantity)
    assert len(spectral_axis) == 1001
    assert spectral_axis.bin_edges[0] == 499*u.AA


def test_create_with_bin_edges():

    wavelengths = np.linspace(500, 2500, 1001) * u.AA
    spectral_axis = SpectralAxis(wavelengths, bin_specification="edges")

    assert np.all(spectral_axis.bin_edges == wavelengths)
    assert spectral_axis[0] == 501*u.AA

    # Test irregular bin edges
    wavelengths = np.array([500, 510, 550, 560, 590])*u.AA
    spectral_axis = SpectralAxis(wavelengths, bin_specification="edges")

    assert np.all(spectral_axis.bin_edges == wavelengths)
    assert np.all(spectral_axis == [505., 530., 555., 575.]*u.AA)


# GENERAL TESTS

# We first run through a series of cases to test different ways of initializing
# the observer and target for SpectralAxis, including for example frames,
# SkyCoords, and making sure that SpectralAxis is not sensitive to the actual
# frame or representation class.

# Local Standard of Rest
LSRD = Galactic(u=0 * u.km, v=0 * u.km, w=0 * u.km,
                U=9 * u.km / u.s, V=12 * u.km / u.s, W=7 * u.km / u.s,
                representation_type='cartesian', differential_type='cartesian')

LSRD_EQUIV = [LSRD,
              SkyCoord(LSRD),  # as a SkyCoord
              LSRD.transform_to(ICRS()),  # different frame
              LSRD.transform_to(ICRS()).transform_to(Galactic())]  # different representation


@pytest.fixture(params=[None] + LSRD_EQUIV)
def observer(request):
    return request.param


# Target located in direction of motion of LSRD with no velocities
LSRD_DIR_STATIONARY = Galactic(u=9 * u.km, v=12 * u.km, w=7 * u.km,
                               representation_type='cartesian')

LSRD_DIR_STATIONARY_EQUIV = [
                             LSRD_DIR_STATIONARY,
                             SkyCoord(LSRD_DIR_STATIONARY),  # as a SkyCoord
                             LSRD_DIR_STATIONARY.transform_to(FK5()),  # different frame
                             LSRD_DIR_STATIONARY.transform_to(ICRS()).transform_to(Galactic())  # different representation
                            ]


@pytest.fixture(params=[None] + LSRD_DIR_STATIONARY_EQUIV)
def target(request):
    return request.param


def test_create_from_spectral_coord(observer, target):
    """
    Checks that parameters are correctly copied from the SpectralCoord object
    to the SpectralAxis object
    """
    spec_coord = SpectralCoord([100, 200, 300] * u.nm, observer=observer,
                               target=target, doppler_convention = 'optical',
                               doppler_rest = 6000*u.AA)
    spec_axis = SpectralAxis(spec_coord)
    assert spec_coord.observer == spec_axis.observer
    assert spec_coord.target == spec_axis.target
    assert spec_coord.radial_velocity == spec_axis.radial_velocity
    assert spec_coord.doppler_convention == spec_axis.doppler_convention
    assert spec_coord.doppler_rest == spec_axis.doppler_rest


def test_create_from_spectral_axis(observer, target):
    """
    Checks that parameters are correctly copied to the new SpectralAxis object
    """
    spec_axis1 = SpectralAxis([100, 200, 300] * u.nm, observer=observer,
                              target=target, doppler_convention = 'optical',
                              doppler_rest = 6000*u.AA)
    spec_axis2 = SpectralAxis(spec_axis1)
    assert spec_axis1.observer == spec_axis2.observer
    assert spec_axis1.target == spec_axis2.target
    assert spec_axis1.radial_velocity == spec_axis2.radial_velocity
    assert spec_axis1.doppler_convention == spec_axis2.doppler_convention
    assert spec_axis1.doppler_rest == spec_axis2.doppler_rest


def test_change_radial_velocity():
    wave = np.linspace(100, 200, 100) * u.AA
    flux = np.ones(100) * u.one
    spec = Spectrum(spectral_axis=wave, flux=flux,
                      radial_velocity=0 * u.km / u.s)

    assert spec.radial_velocity == 0 * u.km/u.s

    with pytest.warns(
        AstropyDeprecationWarning,
        match="Use set_radial_velocity_to or shift_spectrum_to instead"
    ):
        spec.radial_velocity = 1 * u.km / u.s

    assert spec.radial_velocity == 1 * u.km/u.s

    spec = Spectrum(spectral_axis=wave, flux=flux,
                      radial_velocity=10 * u.km / u.s)

    assert spec.radial_velocity == 10 * u.km / u.s

    with pytest.warns(
        AstropyDeprecationWarning,
        match="Use set_radial_velocity_to or shift_spectrum_to instead"
    ):
        spec.radial_velocity = 5 * u.km / u.s

    assert spec.radial_velocity == 5 * u.km / u.s


def test_no_change_radial_velocity():
    wave = np.linspace(100, 200, 100) * u.AA
    flux = np.ones(100) * u.one
    spec = Spectrum(spectral_axis=wave, flux=flux,
                      radial_velocity=0 * u.km / u.s)

    assert spec.radial_velocity == 0 * u.km/u.s
    spec.set_radial_velocity_to(10 * u.km/u.s)
    assert spec.radial_velocity == 10 * u.km/u.s
    assert_quantity_allclose(spec.wavelength, wave)


def test_change_redshift():
    wave = np.linspace(100, 200, 100) * u.AA
    flux = np.ones(100) * u.one
    spec = Spectrum(spectral_axis=wave, flux=flux, redshift=0)

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0))
    assert isinstance(spec.spectral_axis, SpectralAxis)

    with pytest.warns(
        AstropyDeprecationWarning,
        match="Use set_redshift_to or shift_spectrum_to instead"
    ):
        spec.redshift = 0.1

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0.1))
    assert isinstance(spec.spectral_axis, SpectralAxis)

    spec = Spectrum(spectral_axis=wave, flux=flux, redshift=0.2)

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0.2))
    assert isinstance(spec.spectral_axis, SpectralAxis)

    with pytest.warns(
        AstropyDeprecationWarning,
        match="Use set_redshift_to or shift_spectrum_to instead"
    ):
        spec.redshift = 0.4

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0.4))
    assert isinstance(spec.spectral_axis, SpectralAxis)


def test_no_change_redshift():
    wave = np.linspace(100, 200, 100) * u.AA
    flux = np.ones(100) * u.one
    spec = Spectrum(spectral_axis=wave, flux=flux, redshift=0)

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0))
    assert isinstance(spec.spectral_axis, SpectralAxis)

    spec.set_redshift_to(0.5)

    assert spec.redshift.unit.physical_type == 'dimensionless'
    assert_quantity_allclose(spec.redshift, u.Quantity(0.5))
    assert isinstance(spec.spectral_axis, SpectralAxis)

    assert_quantity_allclose(spec.wavelength, wave)


def test_pixel_descending_error():
    '''
    Spectral axes of pixel units must always be ascending.
    This test checks that an error is thrown if one is provided descending
    '''
    flux_unit = u.dimensionless_unscaled
    spec_unit = u.pix

    with pytest.raises(ValueError, match="u.pix spectral axes should always be ascending"):
        Spectrum(spectral_axis=(np.arange(5100, 5300)[::-1])*spec_unit,
                flux=np.random.randn(200)*flux_unit)



# Third-party
import pytest
from astropy.config import set_temp_cache
from astropy.utils.data import download_file

# Package
from specutils.io.default_loaders.apogee import apStar_loader, apVisit_loader, aspcapStar_loader


@pytest.mark.remote_data
@pytest.mark.parametrize(
    "url",
    [
        "https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/stars/apo25m/N7789/apStar-r12-2M00005414+5522241.fits",
        "https://data.sdss.org/sas/dr17/apogee/spectro/redux/dr17/stars/lco25m/293-37-C/asStar-dr17-2M03174068-7748459.fits",
        "https://data.sdss.org/sas/dr13/apogee/spectro/redux/r6/stars/apo25m/4291/apStar-r6-2M07440053+1053592-55960.fits",
    ],
    ids=["apstar_dr16", "asstar_dr17", "apstar_dr13"],
)
def test_apStar_loader(tmp_path, url):
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(url, cache=True)
        spectrum = apStar_loader(filename)  # noqa


@pytest.mark.remote_data
@pytest.mark.parametrize(
    "url",
    [
        "https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/visit/apo25m/N7789/5094/55874/apVisit-r12-5094-55874-123.fits",
        "https://data.sdss.org/sas/dr17/apogee/spectro/redux/dr17/visit/lco25m/293-37-C/11992/59187/asVisit-dr17-11992-59187-227.fits",
        "https://data.sdss.org/sas/dr13/apogee/spectro/redux/r6/apo25m/4925/55702/apVisit-r6-4925-55702-003.fits",
    ],
    ids=["apvisit_dr16", "asvisit_dr17", "apvisit_dr13"],
)
def test_apVisit_loader(tmp_path, url):
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(url, cache=True)
        spectrum = apVisit_loader(filename)  # noqa


@pytest.mark.remote_data
@pytest.mark.parametrize(
    "url",
    [
        "https://data.sdss.org/sas/dr16/apogee/spectro/aspcap/r12/l33/apo25m/N7789/aspcapStar-r12-2M00005414+5522241.fits",
        "https://data.sdss.org/sas/dr17/apogee/spectro/aspcap/dr17/synspec_rev1/lco25m/293-37-C/aspcapStar-dr17-2M03174068-7748459.fits",
    ],
    ids=["aspcap_dr16_apo", "aspcap_dr17_lco"],
)
def test_aspcapStar_loader(tmp_path, url):
    with set_temp_cache(path=str(tmp_path)):
        filename = download_file(url, cache=True)
        spectrum = aspcapStar_loader(filename)  # noqa
